import numpy as np
from scipy.signal import windows

from ..optics import OpticalElement, Apodizer, Wavefront
from ..propagation import FraunhoferPropagator
from ..field import make_focal_grid
from ..fourier import FastFourierTransform, MatrixFourierTransform, FourierFilter

class MultiScaleCoronagraph(OpticalElement):
    '''A phase mask coronagraph.

    This :class:`OpticalElement` simulates the propagation of light through
    a phase mask in the focal plane. To resolve the singularity of the
    phase plate, a multi-scale approach is made. Discretisation errors made at
    a certain level are corrected by the next level with finer sampling.

    Parameters
    ----------
    input_grid : Grid
        The grid on which the incoming wavefront is defined.
    complex_mask : field_generator
        The complex focal-plane phase mask.
    lyot_stop : Field or OpticalElement
        The Lyot stop for the coronagraph. If it's a Field, it is converted to an
        OpticalElement for convenience. If this is None (default), then no Lyot stop is used.
    q : scalar
        The minimum number of pixels per lambda/D. The number of levels in the multi-scale
        Fourier transforms will be chosen to reach at least this number of samples. The required
        q for a high-accuracy vortex coronagraph for example depends on the charge of the vortex. For charge 2,
        this can be as low as 32, but for charge 8 you need ~1024. Lower values give higher performance
        as a smaller number of levels is needed, but increases the sampling errors near the singularity.
        Charges not divisible by four require a much lower q. The default (q=1024) is conservative in
        most cases.
    scaling_factor : scalar
        The fractional increase in spatial frequency sampling per level. Larger scaling factors
        require a smaller number of levels, but each level requires a slower Fourier transform.
        Factors of 2 or 4 usually perform the best.
    window_size : integer
        The size of the next level in the number of pixels on the current layer. Lowering this
        increases performance in exchange for accuracy. Values smaller than 4-8 are not recommended.
    '''
    def __init__(self, input_grid, complex_mask, lyot_stop=None, q=1024, scaling_factor=4, window_size=32):
        self.input_grid = input_grid
        pupil_diameter = input_grid.shape * input_grid.delta

        if hasattr(lyot_stop, 'forward') or lyot_stop is None:
            self.lyot_stop = lyot_stop
        else:
            self.lyot_stop = Apodizer(lyot_stop)

        levels = int(np.ceil(np.log(q / 2) / np.log(scaling_factor))) + 1
        qs = [2 * scaling_factor**i for i in range(levels)]
        num_airys = [input_grid.shape / 2]

        focal_grids = []
        self.focal_masks = []
        self.props = []

        for i in range(1, levels):
            num_airys.append(num_airys[i - 1] * window_size / (2 * qs[i - 1] * num_airys[i - 1]))

        for i in range(levels):
            q = qs[i]
            num_airy = num_airys[i]

            focal_grid = make_focal_grid(q, num_airy, pupil_diameter=pupil_diameter, reference_wavelength=1, focal_length=1)
            focal_mask = complex_mask(focal_grid)

            if i != levels - 1:
                wx = windows.tukey(window_size, 1, False)
                wy = windows.tukey(window_size, 1, False)
                w = np.outer(wy, wx)

                w = np.pad(w, (focal_grid.shape - w.shape) // 2, 'constant').ravel()
                focal_mask *= 1 - w

            for j in range(i):
                fft = FastFourierTransform(focal_grids[j])
                mft = MatrixFourierTransform(focal_grid, fft.output_grid)

                focal_mask -= mft.backward(fft.forward(self.focal_masks[j]))

            if i == 0:
                prop = FourierFilter(input_grid, focal_mask, q)
            else:
                prop = FraunhoferPropagator(input_grid, focal_grid)

            focal_grids.append(focal_grid)
            self.focal_masks.append(focal_mask)
            self.props.append(prop)

    def forward(self, wavefront):
        '''Propagate a wavefront through the coronagraph.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate. This wavefront is expected to be
            in the pupil plane.

        Returns
        -------
        Wavefront
            The Lyot plane wavefront.
        '''
        wavelength = wavefront.wavelength
        wavefront.wavelength = 1

        for i, (mask, prop) in enumerate(zip(self.focal_masks, self.props)):
            if i == 0:
                lyot = Wavefront(prop.forward(wavefront.electric_field), input_stokes_vector=wavefront.input_stokes_vector)
            else:
                focal = prop(wavefront)
                focal.electric_field *= mask
                lyot.electric_field += prop.backward(focal).electric_field

        lyot.wavelength = wavelength
        wavefront.wavelength = wavelength

        if self.lyot_stop is not None:
            lyot = self.lyot_stop.forward(lyot)

        return lyot

    def backward(self, wavefront):
        '''Propagate backwards through the coronagraph.

        This is essentially a forward propagation through the same
        coronagraph, but with the phase pattern inverted.

        Parameters
        ----------
        wavefront : Wavefront
            The Lyot plane wavefront.

        Returns
        -------
        Wavefront
            The pupil-plane wavefront.
        '''
        if self.lyot_stop is not None:
            wavefront = self.lyot_stop.backward(wavefront)

        wavelength = wavefront.wavelength
        wavefront.wavelength = 1

        for i, (mask, prop) in enumerate(zip(self.focal_masks, self.props)):
            if i == 0:
                pup = Wavefront(prop.backward(wavefront.electric_field), input_stokes_vector=wavefront.input_stokes_vector)
            else:
                focal = prop(wavefront)
                focal.electric_field *= mask.conj()
                pup.electric_field += prop.backward(focal).electric_field

        pup.wavelength = wavelength
        wavefront.wavelength = wavelength

        return pup



import numpy as np
import matplotlib.pyplot as plt

from hcipy import *

N = 512

#Defining the grids.
pupil_grid = make_pupil_grid(N)
focal_grid = make_focal_grid(pupil_grid, 8, 5)

#The propagator.
prop = FraunhoferPropagator(pupil_grid, focal_grid)

#Aperture function
aperture = make_circular_aperture(1)(pupil_grid)

#The wavefront in the pupil.
pupil_wavefront = Wavefront(aperture)

#Now we propagate it to the focal plane.
focal_wavefront = prop(pupil_wavefront)

#Setting the amount of photons per unit time.
focal_wavefront.total_power = 1000

#---------------------------------------------------------------------------------
#First example with scalars for the noise sources.
#---------------------------------------------------------------------------------

flatfield = 0.05 # = 5% flat field error
darkcurrentrate = 2 # = dark current counts per second
readnoise = 100 # = rms counts per read out
photonnoise = True

#Creating our detector.
detector_example_1 = NoisyDetector(input_grid=focal_grid, include_photon_noise=photonnoise, flat_field=flatfield, dark_current_rate=darkcurrentrate, read_noise=readnoise)

integration_time = np.logspace(1,6,6)

for t in integration_time:

    #Science measurement.
    detector_example_1.integrate(focal_wavefront, t)
    measurement = detector_example_1.read_out()

    plt.figure()
    imshow_field(measurement)
    plt.colorbar()
    plt.title('Example 1, t = ' + str(t) + ' seconds')

plt.show()

#---------------------------------------------------------------------------------
#Second example with arrays for the noise sources.
#---------------------------------------------------------------------------------

flatfield = 1 + Field(np.random.rand(focal_wavefront.power.size) * 0.05, focal_grid)
darkcurrentrate = Field(np.random.rand(focal_wavefront.power.size) * 2, focal_grid)
readnoise = Field(np.random.rand(focal_wavefront.power.size) * 100, focal_grid)
photonnoise = True

plt.figure()
imshow_field(flatfield)
plt.colorbar()
plt.title('Flat field map')

plt.figure()
imshow_field(darkcurrentrate)
plt.colorbar()
plt.title('Dark current map')

plt.figure()
imshow_field(readnoise)
plt.colorbar()
plt.title('RMS read-noise map')

#Creating our detector.
detector_example_2 = NoisyDetector(input_grid=focal_grid, include_photon_noise=photonnoise, flat_field=flatfield, dark_current_rate=darkcurrentrate, read_noise=readnoise)

integration_time = np.logspace(1,6,6)

for t in integration_time:

    #Science measurement.
    detector_example_2.integrate(focal_wavefront, t)
    measurement = detector_example_2.read_out()

    plt.figure()
    imshow_field(measurement)
    plt.colorbar()
    plt.title('Example 2, t = ' + str(t) + ' seconds')

plt.show()



from hcipy import *
import numpy as np

def test_vortex_coronagraph():
    pupil_grid = make_pupil_grid(256)
    focal_grid = make_focal_grid(4, 32)
    prop = FraunhoferPropagator(pupil_grid, focal_grid)

    aperture = make_circular_aperture(1)
    aperture = evaluate_supersampled(aperture, pupil_grid, 8)

    lyot = make_circular_aperture(0.99)
    lyot = evaluate_supersampled(lyot, pupil_grid, 8) > 1 - 1e-5

    for charge in [2, 4, 6, 8]:
        vortex = VortexCoronagraph(pupil_grid, charge)

        wf = Wavefront(aperture)
        wf.total_power = 1

        img_ref = prop(wf)

        wf = vortex(wf)
        wf.electric_field *= lyot
        img = prop(wf)

        assert img.total_power < 1e-6
        assert img.intensity.max() / img_ref.intensity.max() < 1e-8

def test_fqpm_coronagraph():
    pupil_grid = make_pupil_grid(256)
    focal_grid = make_focal_grid(4, 32)
    prop = FraunhoferPropagator(pupil_grid, focal_grid)

    aperture = make_circular_aperture(1)
    aperture = evaluate_supersampled(aperture, pupil_grid, 8)

    lyot = make_circular_aperture(0.99)
    lyot = evaluate_supersampled(lyot, pupil_grid, 8) > 1 - 1e-5

    fqpm = FQPMCoronagraph(pupil_grid)

    wf = Wavefront(aperture)
    wf.total_power = 1

    img_ref = prop(wf)

    wf = fqpm(wf)
    wf.electric_field *= lyot
    img = prop(wf)

    assert img.total_power < 1e-6
    assert img.intensity.max() / img_ref.intensity.max() < 4e-8

def test_vector_vortex_coronagraph():
    pupil_grid = make_pupil_grid(256)
    focal_grid = make_focal_grid(4, 32)
    prop = FraunhoferPropagator(pupil_grid, focal_grid)

    aperture = make_circular_aperture(1)
    aperture = evaluate_supersampled(aperture, pupil_grid, 8)

    lyot = make_circular_aperture(0.99)
    lyot = evaluate_supersampled(lyot, pupil_grid, 8) > 1 - 1e-5

    for charge in [2, 4, 6, 8]:
        vortex = VectorVortexCoronagraph(charge)

        wf = Wavefront(aperture)
        wf.total_power = 1

        img_ref = prop(wf)

        wf = vortex(wf)
        wf.electric_field *= lyot
        img = prop(wf)

        assert img.total_power < 1e-6
        assert img.intensity.max() / img_ref.intensity.max() < 1e-8

        # Backward check
        # with input 2-tensor electric field
        img.electric_field[0, 1] *= 0
        img.electric_field[1, 0] *= 0
        img.electric_field[1, 1] *= 0
        wf_b = prop.backward(img)
        wf_input_b = vortex.backward(wf_b)

        # with input scalar electric field
        wf_c = prop.backward(Wavefront(img.electric_field[0, 0]))
        wf_input_c = vortex.backward(wf_c)
        assert np.allclose(wf_input_b.electric_field[0, 0], wf_input_c.electric_field[0, 0])

def test_vector_vortex_coronagraph_polarized():
    pupil_grid = make_pupil_grid(256)
    focal_grid = make_focal_grid(4, 32)
    prop = FraunhoferPropagator(pupil_grid, focal_grid)

    aperture = make_obstructed_circular_aperture(1, 0.2)
    aperture = evaluate_supersampled(aperture, pupil_grid, 8)

    lyot = make_obstructed_circular_aperture(0.99, 0.2 / 0.99)
    lyot = evaluate_supersampled(lyot, pupil_grid, 8) > 1 - 1e-5

    stokes_vector = (1, 0, 0, 1)
    for charge in [2]:
        vortex = VectorVortexCoronagraph(charge)

        wf = Wavefront(aperture, input_stokes_vector=stokes_vector)
        wf.total_power = 1

        wf = vortex(wf)
        wf.electric_field *= lyot
        img = prop(wf)

        # incident circular polarization is transfered to orthogonal state
        assert np.allclose(img.intensity, (-1) * img.V)

        # Backward check
        wf_b = prop.backward(img)
        wf_input_b = vortex.backward(wf_b)
        assert np.allclose(wf_input_b.intensity, (-1)**2 * wf_input_b.V)

def test_scalar_vortex_coronagraph_polarized():
    pupil_grid = make_pupil_grid(256)
    focal_grid = make_focal_grid(4, 32)
    prop = FraunhoferPropagator(pupil_grid, focal_grid)

    aperture = make_obstructed_circular_aperture(1, 0.2)
    aperture = evaluate_supersampled(aperture, pupil_grid, 8)

    lyot = make_obstructed_circular_aperture(0.99, 0.2 / 0.99)
    lyot = evaluate_supersampled(lyot, pupil_grid, 8) > 1 - 1e-5

    stokes_vector = (1, 0, 0, 1)
    for charge in [2]:
        vortex = VortexCoronagraph(pupil_grid, charge)

        wf = Wavefront(aperture, input_stokes_vector=stokes_vector)
        wf.total_power = 1

        wf = vortex(wf)
        wf.electric_field *= lyot
        img = prop(wf)

        # polarization state preserved
        assert np.allclose(img.intensity, img.V)

        wf_b = prop.backward(img)
        wf_input_b = vortex.backward(wf_b)
        assert np.allclose(wf_input_b.intensity, wf_input_b.V)

def test_ravc():
    pupil_grid = make_pupil_grid(256)
    focal_grid = make_focal_grid(4, 32)
    prop = FraunhoferPropagator(pupil_grid, focal_grid)

    for co in [0.1, 0.2, 0.3]:
        aperture = lambda grid: make_circular_aperture(1)(grid) - make_circular_aperture(co)(grid)
        aperture = evaluate_supersampled(aperture, pupil_grid, 16)

        aper = aperture > 1e-5

        for charge in [2, 4]:
            pupil, lyot = make_ravc_masks(co, charge, lyot_undersize=0.02)

            pupil = evaluate_supersampled(pupil, pupil_grid, 4)
            lyot = evaluate_supersampled(lyot, pupil_grid, 4)

            vortex = VortexCoronagraph(pupil_grid, charge)

            wf = Wavefront(aper)
            wf.total_power = 1
            img_ref = prop(wf)

            wf = Wavefront(aper)
            wf.total_power = 1
            wf.electric_field *= pupil
            wf = vortex(wf)
            wf.electric_field *= lyot
            img = prop(wf)

            assert img.total_power < 1e-6
            assert img.intensity.max() / img_ref.intensity.max() < 1e-8

            pupil, lyot = make_ravc_masks(co, charge)
            pupil = evaluate_supersampled(pupil, pupil_grid, 4)
            lyot = evaluate_supersampled(lyot, pupil_grid, 4)

            transmission = ((pupil * lyot)**2).sum() / (aperture**2).sum()
            transmission_theoretical = get_ravc_planet_transmission(co, charge)

            assert abs(transmission - transmission_theoretical) < 0.01

def test_app_keller():
    pupil_grid = make_pupil_grid(256)
    focal_grid = make_focal_grid(4, 32)
    propagator = FraunhoferPropagator(pupil_grid, focal_grid)

    aperture = evaluate_supersampled(make_circular_aperture(1), pupil_grid, 8)
    wavefront = Wavefront(aperture)
    wavefront.total_power = 1

    # reference PSF without APP
    img_ref = propagator.forward(wavefront)

    # small rectangular dark zone with 1e-7 contrast
    mask = make_rectangular_aperture(size=(6, 2), center=(9, 0))(focal_grid)
    contrast = 1 - mask + 1e-7

    # APP with aggressive acceleration
    app = generate_app_keller(wavefront, propagator, contrast, num_iterations=150, beta=0.98)
    img = propagator.forward(app)

    assert np.abs(img.intensity.max() / img_ref.intensity.max() - 0.947) < 0.01  # Strehl
    assert np.mean(img.intensity * mask) / np.mean(mask) < 1.6e-8  # contrast

def test_perfect_coronagraph():
    pupil_grid = make_pupil_grid(256)
    aperture = make_circular_aperture(1)(pupil_grid)

    tilts = np.logspace(-3, -1, 51)

    for order in [2, 4, 6, 8]:
        coro = PerfectCoronagraph(aperture, order)

        # Test suppression for on-axis point source
        wf = Wavefront(aperture)
        wf.total_power = 1
        assert coro(wf).total_power < 1e-10

        # Test suppression off-axis
        coronagraph_leakage = []
        for tilt in tilts:
            leakage = coro(Wavefront(aperture * np.exp(2j * np.pi * pupil_grid.x * tilt))).total_power
            coronagraph_leakage.append(leakage)

        y = np.log10(coronagraph_leakage)
        x = np.log10(tilts)
        n = len(x)

        # Do a linear fit on the log-log data to get the power-law coefficient
        beta = ((x * y).sum() - x.sum() * y.sum() / n) / ((x * x).sum() - x.sum()**2 / n)
        assert np.abs(beta - order) / order < 1e-3

def test_lyot_coronagraph():
    pupil_grid = make_pupil_grid(128, 1.1)
    aperture = evaluate_supersampled(make_circular_aperture(1.0), pupil_grid, 8)
    lyot_stop = evaluate_supersampled(make_circular_aperture(0.95), pupil_grid, 8)

    # Coronagraph 1 with the default internal focal length
    fpm_grid = make_focal_grid(q=32, num_airy=3)
    fpm = 1 - evaluate_supersampled(make_circular_aperture(5), fpm_grid, 8)
    cor = LyotCoronagraph(pupil_grid, fpm, lyot_stop)

    # Coronagraph 2 with a large focal length
    focal_length = 10.0
    fpm_grid2 = make_focal_grid(q=32, num_airy=3, spatial_resolution=focal_length)
    fpm2 = 1 - evaluate_supersampled(make_circular_aperture(5 * focal_length), fpm_grid2, 8)
    cor2 = LyotCoronagraph(pupil_grid, fpm2, lyot_stop, focal_length=focal_length)

    # The grid on which the performance is evaluated
    focal_grid = make_focal_grid(q=3, num_airy=25)
    prop = FraunhoferPropagator(pupil_grid, focal_grid)

    wf = Wavefront(aperture)
    wf.total_power = 1
    norm = prop(wf).power.max()
    wf_foc = prop(cor(wf))
    wf_foc2 = prop(cor2(wf))

    # Checks performance of the coronagraph and if the focal length does not introduce artifacts
    assert (wf_foc.power.max() / norm) < 5e-3
    np.testing.assert_allclose(wf_foc.power, wf_foc2.power)


def test_knife_edge_coronagraph():
    grid = make_pupil_grid(64, 1.1)
    aperture = make_circular_aperture(1)(grid)

    focal_grid = make_focal_grid(q=5, num_airy=5)

    prop = FraunhoferPropagator(grid, focal_grid)
    lyot_aperture = make_circular_aperture(0.95)(grid)
    lyot_stop = Apodizer(lyot_aperture)

    wf = Wavefront(aperture)
    wf.total_power = 1.0
    norm = prop(lyot_stop(wf)).power.max()

    directions = ['+x', '-x', '+y', '-y']
    for direction in directions:
        knife_edge = KnifeEdgeLyotCoronagraph(grid, direction=direction, apodizer=None, lyot_stop=lyot_aperture)
        wf_cor = prop(knife_edge(wf))

        assert (wf_cor.power.max() / norm) < 0.3

    directions = ['+x', '-x', '+y', '-y']
    knife_edge_shifts = [-1.0 * grid.x, 1.0 * grid.x, -1.0 * grid.y, 1.0 * grid.y]
    for shift, direction in zip(knife_edge_shifts, directions):
        pre_apodizer = np.exp(1j * 2 * np.pi * shift)
        knife_edge = KnifeEdgeLyotCoronagraph(grid, direction=direction, apodizer=pre_apodizer, lyot_stop=lyot_aperture * np.conj(pre_apodizer))
        wf_cor = prop(knife_edge(wf))

        assert (wf_cor.power.max() / norm) < 1e-2

    # Test for symmetry between a left and right knige edge
    knife_edge_left = KnifeEdgeLyotCoronagraph(grid, direction='+x', lyot_stop=lyot_aperture)
    wf_left = prop(knife_edge_left(wf))

    knife_edge_right = KnifeEdgeLyotCoronagraph(grid, direction='-x', lyot_stop=lyot_aperture)
    wf_right = prop(knife_edge_right(wf))

    # The PSF needs to be flipped and shifted by 1 pixel because the focal grid is odd
    flipped_psf = np.roll(wf_right.power.shaped[:, ::-1], 1, axis=1)
    flipped_psf = Field(flipped_psf.ravel(), focal_grid)

    # Ignore the first column in the evaluation because of the roll over effect.
    assert (abs(wf_left.power.shaped[:, 1::] - flipped_psf.shaped[:, 1::]).max() / norm) < 1e-12

    # Test for symmetry between an up and down knige edge
    knife_edge_down = KnifeEdgeLyotCoronagraph(grid, direction='+y', lyot_stop=lyot_aperture)
    wf_down = prop(knife_edge_down(wf))

    knife_edge_up = KnifeEdgeLyotCoronagraph(grid, direction='-y', lyot_stop=lyot_aperture)
    wf_up = prop(knife_edge_up(wf))

    # The PSF needs to be flipped and shifted by 1 pixel because the focal grid is odd
    flipped_psf = np.roll(wf_up.power.shaped[::-1, :], 1, axis=0)
    flipped_psf = Field(flipped_psf.ravel(), focal_grid)

    # Ignore the first row in the evaluation because of the roll over effect.
    assert (abs(wf_down.power.shaped[1::, :] - flipped_psf.shaped[1::, :]).max() / norm) < 1e-12



import numpy as np
from ..config import Configuration

class Field:
    '''The value of some physical quantity for each point in some coordinate system.

    Parameters
    ----------
    arr : array_like
        An array of values or tensors for each point in the :class:`Grid`.
    grid : Grid
        The corresponding :class:`Grid` on which the values are set.

    Attributes
    ----------
    grid : Grid
        The grid on which the values are defined.
    '''
    def __new__(cls, arr, grid):
        if Configuration().core.use_new_style_fields:
            return super().__new__(NewStyleField)
        else:
            return OldStyleField.__new__(OldStyleField, arr, grid)

    @classmethod
    def from_dict(cls, tree):
        '''Make a Field from a dictionary, previously created by `to_dict()`.

        Parameters
        ----------
        tree : dictionary
            The dictionary from which to make a new Field object.

        Returns
        -------
        Field
            The created object.

        Raises
        ------
        ValueError
            If the dictionary is not formatted correctly.
        '''
        from .grid import Grid

        return cls(np.array(tree['values']), Grid.from_dict(tree['grid']))

    def to_dict(self):
        '''Convert the object to a dictionary for serialization.

        Returns
        -------
        dictionary
            The created dictionary.
        '''
        tree = {
            'values': np.asarray(self),
            'grid': self.grid.to_dict()
        }

        return tree

    def __reduce__(self):
        '''Return a 3-tuple for pickling the Field.

        Returns
        -------
        tuple
            The reduced version of the Field.
        '''
        return (
            _field_reconstruct,
            (self.__class__, np.ndarray, (0,), 'b',),
            self.__getstate__()
        )

    @property
    def tensor_order(self):
        '''The order of the tensor of the field.
        '''
        return self.ndim - 1

    @property
    def tensor_shape(self):
        '''The shape of the tensor of the field.
        '''
        return np.array(self.shape)[:-1]

    @property
    def is_scalar_field(self):
        '''True if this field is a scalar field (ie. a tensor order of 0), False otherwise.
        '''
        return self.tensor_order == 0

    @property
    def is_vector_field(self):
        '''True if this field is a vector field (ie. a tensor order of 1), False otherwise.
        '''
        return self.tensor_order == 1

    @property
    def is_valid_field(self):
        '''True if the field corresponds with its grid.
        '''
        return self.shape[-1] == self.grid.size

    @property
    def shaped(self):
        '''The reshaped version of this field.

        Raises
        ------
        ValueError
            If this field isn't separated, no reshaped version can be made.
        '''
        if not self.grid.is_separated:
            raise ValueError('This field doesn\'t have a shape.')

        if self.tensor_order > 0:
            new_shape = np.concatenate([np.array(self.shape)[:-1], self.grid.shape])
            return self.reshape(new_shape)

        return self.reshape(self.grid.shape)

    def at(self, p):
        '''The value of this field closest to point p.

        Parameters
        ----------
        p : array_like
            The point at which the closest value should be returned.

        Returns
        -------
        array_like
            The value, potentially tensor, closest to point p.
        '''
        i = self.grid.closest_to(p)
        return self[..., i]

class OldStyleField(Field, np.ndarray):
    '''A Field based on subclassing a Numpy array.

    This constitutes an "old-style" Field object. Due to problems and inflexibilities
    in the way subclassing works, we are gradually transitioning to "new-style" fields
    (see :class:`NewStyleField`) which use the dispatch mechanism introduced by
    Numpy 1.13.

    Parameters
    ----------
    arr : array_like
        An array of values or tensors for each point in the :class:`Grid`.
    grid : Grid
        The corresponding :class:`Grid` on which the values are set.

    Attributes
    ----------
    grid : Grid
        The grid on which the values are defined.
    '''
    def __new__(cls, arr, grid):
        obj = np.asarray(arr).view(cls)
        obj.grid = grid
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.grid = getattr(obj, 'grid', None)

    def __getstate__(self):
        '''Get the internal state for pickling.

        Returns
        -------
        tuple
            The state of the Field.
        '''
        data_state = np.ndarray.__reduce__(self)[2]
        return data_state + (self.grid,)

    def __setstate__(self, state):
        '''Set the internal state for pickling.

        Parameters
        ----------
        state : tuple
            The state coming from a __getstate__().
        '''
        _, shp, typ, isf, raw, grid = state

        super().__setstate__((shp, typ, isf, raw))
        self.grid = grid

def _field_reconstruct(subtype, baseclass, baseshape, basetype):
    '''Internal function for building a new Field object for pickling.

    Parameters
    ----------
    subtype
        The class of Field.
    baseclass
        The array class that was used for the Field.
    baseshape
        The shape of the Field.
    basetype
        The data type of the Field.

    Returns
    -------
    Field
        The built Field object.
    '''
    data = np.ndarray.__new__(baseclass, baseshape, basetype)
    grid = None

    return subtype.__new__(subtype, data, grid)

def _unwrap(arg):
    if isinstance(arg, Field):
        return arg.data

    if isinstance(arg, list):
        return [_unwrap(x) for x in arg]

    if isinstance(arg, dict):
        return {key: _unwrap(val) for key, val in arg.items()}

    if isinstance(arg, tuple):
        return tuple(_unwrap(x) for x in arg)

    return arg

class NewStyleField(Field, np.lib.mixins.NDArrayOperatorsMixin):
    '''A Field based on composition rather than subclassing a Numpy array.

    This constitutes an "new-style" Field object. A previous version uses subclassing
    from Numpy arrays (see :class:`OldStyleField` that had problems and inflexibilities
    due to the way subclassing works. We are gradually transitioning to "new-style" fields,
    which use the dispatch mechanism introduced by Numpy 1.13.

    Parameters
    ----------
    data : array_like
        An array of values or tensors for each point in the :class:`Grid`.
    grid : Grid
        The corresponding :class:`Grid` on which the values are set.

    Attributes
    ----------
    data : array_like
        The underlying data as a Numpy array or compatible object.
    grid : Grid
        The grid on which the values are defined.
    '''
    def __init__(self, data, grid):
        self.data = np.asarray(data)
        self.grid = grid

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        out = kwargs.get('out', ())

        inputs = _unwrap(inputs)

        if out:
            kwargs['out'] = tuple(x.data if is_field(x) else x for x in out)

        result = getattr(ufunc, method)(*inputs, **kwargs)

        # Note: this is an extremely simple way of determining whether the
        # resulting object should be a Field or not. We can likely do better
        # by looking directly at the name of the function that we are executing.
        # The current code is fine for most ufuncs.
        if isinstance(result, np.ndarray):
            return Field(result, self.grid)
        else:
            return result

    def __array__(self, dtype=None):
        if dtype is None:
            return self.data

        return self.data.astype(dtype, copy=False)

    def __array_function__(self, func, types, args, kwargs):
        args = _unwrap(args)

        result = func(*args, **kwargs)

        # Note: this is an extremely simple way of determining whether the
        # resulting object should be a Field or not. We can likely do better
        # by looking directly at the name of the function that we are executing.
        if isinstance(result, np.ndarray):
            return Field(result, self.grid)

        return result

    def __getitem__(self, indices):
        res = self.data[indices]
        if np.isscalar(res):
            return res

        return Field(res, self.grid)

    def __setitem__(self, indices, values):
        self.data[indices] = values

        return self

    def __getstate__(self):
        '''Get the internal state for pickling.

        Returns
        -------
        tuple
            The state of the Field.
        '''
        data_state = self.data.__reduce__()[2]
        return data_state + (self.grid,)

    def __setstate__(self, state):
        '''Set the internal state for pickling.

        Parameters
        ----------
        state : tuple
            The state coming from a __getstate__().
        '''
        _, shp, typ, isf, raw, grid = state

        self.data = np.array([])
        self.data.__setstate__((shp, typ, isf, raw))
        self.grid = grid

    @property
    def T(self):
        return np.transpose(self)

    @property
    def dtype(self):
        return self.data.dtype

    @property
    def imag(self):
        return np.imag(self)

    @property
    def real(self):
        return np.real(self)

    @property
    def size(self):
        return np.prod(self.shape)

    @property
    def ndim(self):
        return len(self.shape)

    @property
    def shape(self):
        return self.data.shape

    def astype(self, dtype, *args, **kwargs):
        return Field(self.data.astype(dtype, *args, **kwargs), self.grid)

    def __len__(self):
        return len(self.data)

    def conj(self, *args, **kwargs):
        return np.conj(self, *args, **kwargs)

    def conjugate(self, *args, **kwargs):
        return np.conjugate(self, *args, **kwargs)

    @property
    def flags(self):
        return self.data.flags

    all = np.all
    any = np.any
    argmax = np.argmax
    argmin = np.argmin
    argpartition = np.argpartition
    argsort = np.argsort
    clip = np.clip
    compress = np.compress
    copy = np.copy
    cumprod = np.cumprod
    cumsum = np.cumsum
    dot = np.dot
    flatten = np.ravel
    max = np.max
    mean = np.mean
    min = np.min
    nonzero = np.nonzero
    prod = np.prod
    ptp = np.ptp
    ravel = np.ravel
    repeat = np.repeat
    reshape = np.reshape
    round = np.round
    sort = np.sort
    squeeze = np.squeeze
    std = np.std
    sum = np.sum
    trace = np.trace
    transpose = np.transpose
    var = np.var

def is_field(obj):
    return isinstance(obj, Field)



import numpy as np
from ..mode_basis import ModeBasis
from ..field import Field
from .deformable_mirror import DeformableMirror

class TipTiltMirror(DeformableMirror):
    '''A tip-tilt mirror.

    This deformable mirror class provides a tip-tilt mirror, where the
    actuators are controlled in OPD.

    Parameters
    ----------
    input_grid : Grid
        The grid that corresponds to the input wavefront
    '''
    def __init__(self, input_grid):
        self.input_grid = input_grid
        self.actuators = np.zeros((2,))

        modes = ModeBasis([Field(self.input_grid.x, self.input_grid), Field(self.input_grid.y, self.input_grid)])

        super(TipTiltMirror, self).__init__(modes)



import numpy as np
from hcipy import *
import scipy.special
import pytest

def test_fraunhofer_propagation_circular():
    for num_pix in [512, 1024]:
        pupil_grid = make_pupil_grid(num_pix)
        focal_grid = make_focal_grid(16, 8)

        ind = focal_grid.closest_to((0, 0))

        for diameter in [1, 0.7]:
            aperture = evaluate_supersampled(make_circular_aperture(diameter), pupil_grid, 8)

            for focal_length in [1, 0.8]:
                prop = FraunhoferPropagator(pupil_grid, focal_grid, focal_length=focal_length)

                for wavelength in [1, 0.5]:
                    wf = Wavefront(aperture, wavelength)
                    img = prop(wf).electric_field
                    img /= img[np.argmax(np.abs(img))]

                    x = focal_grid.as_('polar').r * np.pi / wavelength * diameter / focal_length
                    x[ind] = 1
                    reference = 2 * scipy.special.jv(1, x) / x
                    reference[ind] = 1

                    if num_pix == 512:
                        assert np.abs(img - reference).max() < 3e-5
                    elif num_pix == 1024:
                        assert np.abs(img - reference).max() < 1e-5
                    else:
                        # This should never happen.
                        assert False

def test_fraunhofer_propagation_rectangular():
    for num_pix in [512, 1024]:
        pupil_grid = make_pupil_grid(num_pix)
        focal_grid = make_focal_grid(16, 8)

        for size in [[1, 1], [0.75, 1], [0.75, 0.75]]:
            aperture = evaluate_supersampled(make_rectangular_aperture(size), pupil_grid, 8)

            for focal_length in [1, 1.3]:
                prop = FraunhoferPropagator(pupil_grid, focal_grid, focal_length=focal_length)

                for wavelength in [1, 0.8]:
                    wf = Wavefront(aperture, wavelength)
                    img = prop(wf).electric_field
                    img /= img[np.argmax(np.abs(img))]

                    k_x, k_y = np.array(size) / wavelength / focal_length
                    reference = (np.sinc(k_x * focal_grid.x) * np.sinc(k_y * focal_grid.y))

                    if num_pix == 512:
                        assert np.abs(img - reference).max() < 5e-5
                    elif num_pix == 1024:
                        assert np.abs(img - reference).max() < 2e-5
                    else:
                        # This should never happen.
                        assert False

def single_propagation_test(propagator, number_of_pixels, wavelength, a, b, relative_distance, error_threshold):
    wavenumber = 2 * np.pi / wavelength

    pupil_grid = make_pupil_grid(number_of_pixels, [16 * a, 16 * b])
    threshold_distance = np.min(pupil_grid.delta * np.array([16 * a, 16 * b]) / wavelength)
    distance = relative_distance * threshold_distance

    prop_pos = propagator(pupil_grid, distance, num_oversampling=2)
    prop_neg = propagator(pupil_grid, -distance, num_oversampling=2)

    aperture = evaluate_supersampled(make_rectangular_aperture([2 * a, 2 * b]), pupil_grid, 2)

    img_forward = prop_pos.forward(Wavefront(aperture, wavelength)).intensity
    img_backward = prop_pos.backward(Wavefront(aperture, wavelength)).intensity

    img_forward_neg = prop_neg.forward(Wavefront(aperture, wavelength)).intensity
    img_backward_neg = prop_neg.forward(Wavefront(aperture, wavelength)).intensity

    assert np.allclose(img_forward, img_backward_neg)
    assert np.allclose(img_backward, img_forward_neg)

    def generate_reference_field(reference_grid):
        w_x1 = np.sqrt(2 / (distance * wavelength)) * (reference_grid.x - a)
        w_x2 = np.sqrt(2 / (distance * wavelength)) * (reference_grid.x + a)
        w_y1 = np.sqrt(2 / (distance * wavelength)) * (reference_grid.y - b)
        w_y2 = np.sqrt(2 / (distance * wavelength)) * (reference_grid.y + b)

        fresnel = scipy.special.fresnel

        ssa_x1, csa_x1 = fresnel(w_x1)
        ssa_x2, csa_x2 = fresnel(w_x2)
        ssa_y1, csa_y1 = fresnel(w_y1)
        ssa_y2, csa_y2 = fresnel(w_y2)

        F_x1 = csa_x1 + 1j * ssa_x1
        F_x2 = csa_x2 + 1j * ssa_x2
        F_y1 = csa_y1 + 1j * ssa_y1
        F_y2 = csa_y2 + 1j * ssa_y2

        reference = (F_x2 - F_x1) * (F_y2 - F_y1)
        reference *= -1j * np.exp(1j * wavenumber * distance) / 2
        reference *= np.exp(1j * wavenumber * (reference_grid.x**2 + reference_grid.y**2) / (2 * distance))
        reference = np.abs(Field(reference, reference_grid))**2

        return reference

    reference_image = evaluate_supersampled(generate_reference_field, pupil_grid, 2)

    img_forward = subsample_field(img_forward, 8)
    img_backward = subsample_field(img_backward, 8)
    reference_image = subsample_field(reference_image, 8)

    absolute_error_forward = np.abs(img_forward - reference_image).max()
    assert absolute_error_forward < error_threshold

    absolute_error_backward = np.abs(img_backward - reference_image).max()
    assert absolute_error_backward < error_threshold

def test_fresnel_propagation_rectangular_quick():
    for num_pix in [512]:
        for wavelength in [500e-9]:
            for a, b in [[0.001, 0.001]]:
                for relative_distance, error_threshold in zip([0.2, 5], [0.02, 0.01]):
                    single_propagation_test(FresnelPropagator, num_pix, wavelength, a, b, relative_distance, error_threshold)

def test_asp_propagation_rectangular_quick():
    for num_pix in [512]:
        for wavelength in [500e-9]:
            for a, b in [[0.001, 0.001]]:
                for relative_distance, error_threshold in zip([0.2, 5], [0.02, 0.01]):
                    single_propagation_test(AngularSpectrumPropagator, num_pix, wavelength, a, b, relative_distance, error_threshold)

@pytest.mark.slow
def test_fresnel_propagation_rectangular():
    for num_pix in [512, 1024]:
        for wavelength in [500e-9, 700e-9]:
            for a, b in [[0.001, 0.001], [0.0015, 0.001]]:
                for relative_distance, error_threshold in zip([0.2, 1, 5], [0.02, 0.01, 0.01]):
                    single_propagation_test(FresnelPropagator, num_pix, wavelength, a, b, relative_distance, error_threshold)

@pytest.mark.slow
def test_asp_propagation_rectangular():
    for num_pix in [512, 1024]:
        for wavelength in [500e-9, 700e-9]:
            for a, b in [[0.001, 0.001], [0.0015, 0.001]]:
                for relative_distance, error_threshold in zip([0.2, 1, 5], [0.02, 0.01, 0.01]):
                    single_propagation_test(AngularSpectrumPropagator, num_pix, wavelength, a, b, relative_distance, error_threshold)



import numpy as np
from .detector import Detector
from .optical_element import OpticalElement, AgnosticOpticalElement, make_agnostic_forward, make_agnostic_backward
from .wavefront import Wavefront
from ..mode_basis import ModeBasis, make_lp_modes
from ..field import Field, CartesianGrid, RegularCoords
from ..dev import deprecated

def make_gaussian_fiber_mode(mode_field_diameter):
    '''The Gaussian approximation of a fiber mode.

    Parameters
    ----------
    mode_field_diameter : scalar
        The mode field diameter of the fiber.

    Returns
    -------
    Field generator
        The Gaussian fiber mode.
    '''
    def func(grid):
        if grid.is_('cartesian'):
            r2 = grid.x**2 + grid.y**2
        else:
            r2 = grid.as_('polar').r**2

        # Compute and normalize the Gaussian function.
        mode_field_radius = 0.5 * mode_field_diameter
        res = np.exp(-r2 / mode_field_radius**2)
        res *= np.sqrt(2 / np.pi) / mode_field_radius

        return Field(res, grid)

    return func

class StepIndexFiber(AgnosticOpticalElement):
    '''A step-index fiber.

    The modal behaviour of the step-index fiber depends on the input parameters.

    Parameters
    ----------
    core_radius : scalar
        The radius of the high-index core.
    NA : scalar or function of wavelength
        The numerical aperture of the fiber.
    fiber_length : scalar
        The length of the optical fiber.
    position : array_like
        The 2D position of the fiber in the optical plane.
    '''
    def __init__(self, core_radius, NA, fiber_length, position=None):
        super().__init__(False, True)

        self._core_radius = core_radius
        self._NA = NA
        self.fiber_length = fiber_length

        self._position = np.zeros((2,)) if position is None else np.array(position)

    def make_instance(self, instance_data, input_grid, output_grid, wavelength):
        monochromatic_V = self.V(wavelength)

        instance_data.NA = self.evaluate_parameter(self._NA, input_grid, output_grid, wavelength)
        instance_data.fiber_modes, instance_data.beta = make_lp_modes(input_grid.shifted(-self._position), monochromatic_V, self.core_radius, return_betas=True)

    def num_modes(self, wavelength):
        '''The approximate number of modes of the fiber.
        '''
        V = self.V(wavelength)
        return V**2 / 2

    def V(self, wavelength):  # noqa: N802
        '''The normalized frequency parameter for step-index fibers.
        '''
        return 2 * np.pi / wavelength * self.core_radius * self.NA

    def mode_field_radius(self, wavelength):
        '''The mode field radius of the fiber.

        The mode field radius is the radius of the gaussian beam best matched to the fundamental mode [Marcuse1977]_.

        .. [Marcuse1977] D. Marcuse 1977, "Loss analysis of single-mode fiber splices," The Bell System Technical Journal 56, 703-718 (2014)
        '''
        V = self.V(wavelength)
        w = self.core_radius * (0.65 + 1.619 / V**(3 / 2) + 2.879 / V**6)
        return w

    @property
    def core_radius(self):
        '''The core radius of this fiber.
        '''
        return self._core_radius

    @core_radius.setter
    def core_radius(self, core_radius):
        self._core_radius = core_radius
        self.clear_cache()

    @property
    def position(self):
        '''The 2D position of the fiber.
        '''
        return self._position

    @position.setter
    def position(self, position):
        self._position = np.array(position)
        self.clear_cache()

    @property
    def NA(self):  # noqa: N802
        '''The numerical aperture of this fiber.
        '''
        return self._NA

    @NA.setter
    def NA(self, NA):  # noqa: N802
        self._NA = NA
        self.clear_cache()

    numerical_aperture = NA

    def get_input_grid(self, output_grid, wavelength):
        return output_grid

    def get_output_grid(self, input_grid, wavelength):
        return input_grid

    @make_agnostic_forward
    def modal_decomposition(self, instance_data, wavefront):
        '''Decompose the input wavefront into the modal distribution of the fiber.

        Parameters
        ----------
        wavefront : Wavefront
            The incoming wavefront.

        Returns
        -------
        array_like
            The modal coefficients.
        '''

        M = instance_data.fiber_modes.transformation_matrix
        mode_coefficients = np.einsum('...i, ij->...j', wavefront.electric_field * wavefront.grid.weights, M.conj())
        return mode_coefficients

    @make_agnostic_forward
    def forward(self, instance_data, wavefront):
        '''Forward propagate the light through the fiber.

        Parameters
        ----------
        wavefront : Wavefront
            The incoming wavefront.

        Returns
        -------
        Wavefront
            The wavefront that exits the fiber.
        '''
        M = instance_data.fiber_modes.transformation_matrix
        mode_coefficients = np.einsum('...i, ij->...j', wavefront.electric_field * wavefront.grid.weights, M.conj())
        output_electric_field = np.einsum('...i, ij->...j', mode_coefficients * np.exp(1j * instance_data.beta * self.fiber_length), M.T.conj())
        output_electric_field = Field(output_electric_field, wavefront.grid)

        return Wavefront(output_electric_field, wavefront.wavelength)

    @make_agnostic_backward
    def backward(self, instance_data, wavefront):
        '''Backward propagate the light through the fiber.

        Parameters
        ----------
        wavefront : Wavefront
            The incoming wavefront.

        Returns
        -------
        Wavefront
            The wavefront that exits the fiber.
        '''
        M = instance_data.fiber_modes.transformation_matrix
        mode_coefficients = np.einsum('...i, ij->...j', wavefront.electric_field * wavefront.grid.weights, M.conj())
        output_electric_field = np.einsum('...i, ij->...j', mode_coefficients * np.exp(-1j * instance_data.beta * self.fiber_length), M.T.conj())
        output_electric_field = Field(output_electric_field, wavefront.grid)

        return Wavefront(output_electric_field, wavefront.wavelength)

@deprecated('This class operated counter to HCIPy best practices and will be removed in HCIPy 0.7 or later. Please use a combination of the SingleModeFiberInjection and Detector classes instead.')
class SingleModeFiber(Detector):
    def __init__(self, input_grid, mode_field_diameter, mode=None):
        self.input_grid = input_grid
        self.mode_field_diameter = mode_field_diameter

        if mode is None:
            mode = make_gaussian_fiber_mode(mode_field_diameter)

        try:
            self.mode = mode(self.input_grid, mode_field_diameter)
        except TypeError:
            self.mode = mode(self.input_grid)

        self.mode /= np.sqrt(np.sum(np.abs(self.mode)**2 * self.input_grid.weights))
        self.intensity = 0

    def integrate(self, wavefront, dt, weight=1):
        self.intensity += weight * dt * np.abs(np.dot(wavefront.electric_field.conj() * wavefront.electric_field.grid.weights, self.mode))**2

    def read_out(self):
        intensity = self.intensity
        self.intensity = 0
        return intensity

class SingleModeFiberInjection(OpticalElement):
    '''Injection into a single mode fiber.

    Parameters
    ----------
    input_grid : Grid
        The grid of the incoming wavefront.
    mode : Field generator
        The mode of the fiber.
    position : array_like
        The 2D position of the fiber. If this is None, the fiber will be located at (0, 0).
    '''
    def __init__(self, input_grid, mode, position=None):
        self.input_grid = input_grid

        if position is None:
            position = np.zeros(2)
        else:
            position = np.array(position)

        self.output_grid = CartesianGrid(RegularCoords([1, 1], [1, 1], position))
        self.output_grid.weights = 1

        # Compute the fiber mode and normalize.
        self.mode = mode(self.input_grid.shifted(-position))
        self.mode.grid = self.input_grid
        self.mode /= np.sqrt(np.sum(np.abs(self.mode)**2 * self.input_grid.weights))

    def forward(self, wavefront):
        '''Forward propagate the light through the fiber.

        Parameters
        ----------
        wavefront : Wavefront
            The incoming wavefront.

        Returns
        -------
        Wavefront
            The complex amplitude of the single-mode fiber.
        '''
        output = np.dot(wavefront.electric_field.conj() * self.input_grid.weights, self.mode)

        if not np.isscalar(output):
            output = np.array(output)
        else:
            # The incoming wavefront was polarized, so we need to add a dimension to
            # make it a valid Field.
            output = output[..., np.newaxis]

        output = Field(output, self.output_grid)

        return Wavefront(output, wavefront.wavelength, wavefront.input_stokes_vector)

    def backward(self, wavefront):
        '''Backwards propagate the light through the fiber.

        Parameters
        ----------
        wavefront : Wavefront
            The complex amplitude for the single-mode fiber.

        Returns
        -------
        Wavefront
            The outgoing wavefront.
        '''
        output = wavefront.electric_field * self.mode
        output = Field(output, self.input_grid)

        return Wavefront(output, wavefront.wavelength, wavefront.input_stokes_vector)

class SingleModeFiberArray(OpticalElement):
    def __init__(self, input_grid, fiber_grid, mode, *args, **kwargs):
        '''An array of single-mode fibers.

        This implementation assumes orthogonality of the different fibers. It
        does not attempt to perform any orthogonalization of the fibers. This
        may lead to creation of light power.

        Additionally, all fiber modes are assumed to be independent of wavelength.

        Parameters
        ----------
        input_grid : Grid
            The grid on which the input wavefront is defined.
        fiber_grid : Grid
            The centers of each of the single-mode fibers.
        mode : function
            The mode of the single-mode fibers. The function should take a grid
            and return the amplitude of the fiber mode.
        '''
        self.input_grid = input_grid
        self.fiber_grid = fiber_grid

        self.fiber_modes = [mode(input_grid.shifted(-p), *args, **kwargs) for p in fiber_grid]
        self.fiber_modes = [m / np.sqrt(np.sum(np.abs(m)**2 * input_grid.weights)) for m in self.fiber_modes]
        self.fiber_modes = ModeBasis(self.fiber_modes)

        self.projection_matrix = self.fiber_modes.transformation_matrix

    def forward(self, wavefront):
        '''Forward propagate the light through the fiber array.

        Parameters
        ----------
        wavefront : Wavefront
            The incoming wavefront.

        Returns
        -------
        Wavefront
            The complex amplitudes of each fiber.
        '''
        res = self.projection_matrix.T.dot(wavefront.electric_field.conj() * self.input_grid.weights)
        return Wavefront(Field(res, self.fiber_grid), wavefront.wavelength)

    def backward(self, wavefront):
        '''Backwards propagate the light through the fiber array.

        Parameters
        ----------
        wavefront : Wavefront
            The complex amplitudes for each of the single-mode fibers.

        Returns
        -------
        Wavefront
            The outgoing wavefront.
        '''
        res = self.projection_matrix.dot(wavefront.electric_field)
        return Wavefront(Field(res, self.input_grid), wavefront.wavelength)



from hcipy import *
import numpy as np
import pytest
from packaging import version
import scipy.signal

def make_all_fourier_transforms(input_grid, q, fov, shift):
    fft1 = FastFourierTransform(input_grid, q=q, fov=fov, shift=shift, emulate_fftshifts=True)
    fft2 = FastFourierTransform(input_grid, q=q, fov=fov, shift=shift, emulate_fftshifts=False)
    mft1 = MatrixFourierTransform(input_grid, fft1.output_grid, precompute_matrices=True, allocate_intermediate=True)
    mft2 = MatrixFourierTransform(input_grid, fft1.output_grid, precompute_matrices=True, allocate_intermediate=False)
    mft3 = MatrixFourierTransform(input_grid, fft1.output_grid, precompute_matrices=False, allocate_intermediate=True)
    mft4 = MatrixFourierTransform(input_grid, fft1.output_grid, precompute_matrices=False, allocate_intermediate=False)
    nft1 = NaiveFourierTransform(input_grid, fft1.output_grid, precompute_matrices=True)
    nft2 = NaiveFourierTransform(input_grid, fft1.output_grid, precompute_matrices=False)
    zfft = ZoomFastFourierTransform(input_grid, fft1.output_grid)

    return [fft1, fft2, mft1, mft2, mft3, mft4, nft1, nft2, zfft]

def check_energy_conservation(dtype, shift_input, scale, shift_output, q, fov, dims):
    grid = make_uniform_grid(dims, 1, has_center=True).shifted(shift_input).scaled(scale)
    f_in = Field(np.random.randn(grid.size), grid).astype(dtype)

    energy_in = np.sum(np.abs(f_in)**2 * f_in.grid.weights)

    fourier_transforms = make_all_fourier_transforms(grid, q, fov, shift_output)

    energy_ratios = []
    patterns_match = []
    for ft1 in fourier_transforms:
        f_inter = ft1.forward(f_in)
        assert f_inter.dtype == np.dtype(dtype)

        for ft2 in fourier_transforms:
            f_out = ft2.backward(f_inter)
            assert f_out.dtype == np.dtype(dtype)

            energy_out = np.sum(np.abs(f_out)**2 * f_out.grid.weights)
            energy_ratio = energy_out / energy_in

            pattern_match = np.std(np.abs(f_out - f_in)) / np.abs(f_in).mean()

            energy_ratios.append(energy_ratio)
            patterns_match.append(pattern_match)

    energy_ratios = np.array(energy_ratios).reshape((len(fourier_transforms), len(fourier_transforms)))
    patterns_match = np.array(patterns_match).reshape((len(fourier_transforms), len(fourier_transforms)))

    if fov == 1:
        # When the full fov is retained, the pattern should be the same and energy should
        # be conserved. We use different accuracy limits based on bit depth.
        if np.dtype(dtype) == np.dtype('complex128'):
            assert np.all(patterns_match < 1e-12)
            assert np.all(np.abs(energy_ratios - 1) < 1e-12)
        else:
            assert np.all(patterns_match < 1e-6)
            assert np.all(np.abs(energy_ratios - 1) < 1e-5)
    else:
        # If the full fov is not retained, the pattern and energy loss should be the same
        # for all fourier transform combinations.
        assert np.allclose(energy_ratios, energy_ratios[0, 0])
        assert np.allclose(patterns_match, patterns_match[0, 0])

@pytest.mark.parametrize('dtype', ['complex128', 'complex64'])
def test_fourier_energy_conservation_1d(dtype):
    np.random.seed(0)

    for shift_input in [0, 0.1]:
        for scale in [1, 2]:
            for shift_output in [0, 0.1]:
                for q in [1, 1.23, 3, 4]:
                    for fov in [1, 0.5, 0.8]:
                        for dims in [64, 65]:
                            check_energy_conservation(dtype, shift_input, scale, shift_output, q, fov, dims)

@pytest.mark.parametrize('dtype', ['complex128', 'complex64'])
def test_fourier_energy_conservation_2d(dtype):
    np.random.seed(0)

    for shift_input in [[0, 0], [0.1]]:
        for scale in [1, 2]:
            for shift_output in [[0, 0], [0.1]]:
                for q in [1, 1.23, 3, 4]:
                    for fov in [1, 0.5, 0.8]:
                        for dims in [[8, 8], [8, 16], [9, 9], [9, 18]]:
                            check_energy_conservation(dtype, shift_input, scale, shift_output, q, fov, dims)

def check_symmetry(dtype, scale, shift_output, q, fov, dims):
    tol = 1e-12 if dtype == 'complex128' else 1e-6

    input_grid = make_uniform_grid(dims, 1, has_center=False).scaled(scale)

    # Check symmetry of the input grid.
    for i in range(input_grid.ndim):
        assert np.allclose(input_grid.coords[i], -input_grid.coords[i][::-1])

    # Make even and odd input fields.
    f_in = Field(np.random.randn(input_grid.size), input_grid).astype(dtype)

    f_in_even = (f_in + f_in[::-1]) / 2
    f_in_odd = (f_in - f_in[::-1]) / 2

    assert np.allclose(f_in, f_in_even + f_in_odd, atol=tol)

    # Make different Fourier transforms.
    fourier_transforms = make_all_fourier_transforms(input_grid, q, fov, shift_output)

    for ft in fourier_transforms:
        # Check even and odd Fourier symmetries.
        ft_even = ft.forward(f_in_even)
        ft_odd = ft.forward(f_in_odd)

        assert np.allclose(ft_even.imag, 0, atol=tol)
        assert np.allclose(ft_odd.real, 0, atol=tol)

        # Check linearity.
        ft_both = ft.forward(f_in)
        assert np.allclose(ft_both, ft_even + ft_odd, atol=tol)

        ft_both_double = ft.forward(f_in * 2)
        assert np.allclose(ft_both_double, ft_both * 2, atol=tol)

@pytest.mark.parametrize('dtype', ['complex128', 'complex64'])
def test_fourier_symmetries_1d(dtype):
    np.random.seed(0)

    for scale in [1, 2]:
        for shift_output in [0, 0.1]:
            for q in [1, 1.23, 3, 4]:
                for fov in [1, 0.5, 0.8]:
                    for dims in [64, 65]:
                        check_symmetry(dtype, scale, shift_output, q, fov, dims)

@pytest.mark.parametrize('dtype', ['complex128', 'complex64'])
def test_fourier_symmetries_2d(dtype):
    np.random.seed(0)

    for scale in [1, 2]:
        for shift_output in [[0, 0], [0.1]]:
            for q in [1, 1.23, 3, 4]:
                for fov in [1, 0.5, 0.8]:
                    for dims in [[8, 8], [8, 16], [9, 9], [9, 18]]:
                        check_symmetry(dtype, scale, shift_output, q, fov, dims)

def test_make_fourier_transform():
    input_grid = make_pupil_grid(128)

    ft = make_fourier_transform(input_grid, q=1, fov=1, shift=0.1, planner='estimate')
    assert type(ft) == FastFourierTransform
    assert ft.input_grid == input_grid
    assert ft.output_grid == make_fft_grid(input_grid, q=1, fov=1, shift=0.1)

    fft_grid = make_fft_grid(input_grid, q=1, fov=1, shift=0.1)
    ft = make_fourier_transform(input_grid, fft_grid, planner='estimate')
    assert type(ft) == FastFourierTransform
    assert ft.input_grid == input_grid
    assert ft.output_grid == fft_grid

    ft = make_fourier_transform(input_grid, q=8, fov=0.3, shift=0.1, planner='estimate')
    assert type(ft) == MatrixFourierTransform
    assert ft.input_grid == input_grid
    assert ft.output_grid == make_fft_grid(input_grid, q=8, fov=0.3, shift=0.1)

    fft_grid = make_fft_grid(input_grid, q=8, fov=0.3, shift=0.1)
    ft = make_fourier_transform(input_grid, fft_grid, planner='estimate')
    assert type(ft) == MatrixFourierTransform
    assert ft.input_grid == input_grid
    assert ft.output_grid == fft_grid

    ft = make_fourier_transform(input_grid, q=1, fov=1, shift=0.1, planner='measure')
    ft = make_fourier_transform(input_grid, q=8, fov=0.1, shift=0.1, planner='measure')

    output_grid = CartesianGrid(UnstructuredCoords([np.random.randn(100), np.random.randn(100)]))
    ft = make_fourier_transform(input_grid, output_grid)
    assert type(ft) == NaiveFourierTransform
    assert ft.input_grid == input_grid
    assert ft.output_grid == output_grid

def test_fft_grid_reconstruction():
    for shift_input in [[0, 0], [0.1]]:
            for scale in [1, 2]:
                for shift_output in [[0, 0], [0.1]]:
                    for q in [1, 1.234, 3, 4]:
                        for fov in [1, 0.5, 0.8, [0.3, 0.23]]:
                            for dims in [[8, 8], [8, 16], [9, 9], [9, 18]]:
                                input_grid = make_uniform_grid(dims, 1, has_center=True).shifted(shift_input).scaled(scale)

                                fft_grid = make_fft_grid(input_grid, q, fov, shift_output)

                                assert is_fft_grid(fft_grid, input_grid)

                                q_recon, fov_recon, shift_output_recon = get_fft_parameters(fft_grid, input_grid)
                                fft_grid_recon = make_fft_grid(input_grid, q_recon, fov_recon, shift_output_recon)

                                print(q, fov, shift_output)
                                print(q_recon, fov_recon, shift_output_recon)

                                assert np.allclose(fft_grid.x, fft_grid_recon.x)
                                assert np.allclose(fft_grid.y, fft_grid_recon.y)

    # Check raising behaviour.
    input_grid = make_uniform_grid([128, 128], 1)
    output_grid = CartesianGrid(SeparatedCoords(input_grid.separated_coords))

    with pytest.raises(ValueError):
        get_fft_parameters(output_grid, input_grid)
    with pytest.raises(ValueError):
        get_fft_parameters(input_grid, output_grid)

    assert not is_fft_grid(output_grid, input_grid)
    assert not is_fft_grid(input_grid, output_grid)

    output_grid = make_fft_grid(input_grid, 1, 0.5).scaled(1.001)

    with pytest.raises(ValueError):
        get_fft_parameters(output_grid, input_grid)

    assert not is_fft_grid(output_grid, input_grid)

def test_fft_exceptions():
    regular_grid = make_uniform_grid([128, 128], 1)
    irregular_cartesian_grid = CartesianGrid(UnstructuredCoords([np.random.randn(1024), np.random.randn(1024)]))
    irregular_polar_grid = irregular_cartesian_grid.as_('polar')

    with pytest.raises(ValueError):
        FastFourierTransform(regular_grid, 0.9, 0.5)

    with pytest.raises(ValueError):
        FastFourierTransform(regular_grid, 2, 1.1)

    with pytest.raises(ValueError):
        FastFourierTransform(regular_grid, 2, -0.1)

    with pytest.raises(ValueError):
        FastFourierTransform(irregular_cartesian_grid, 2, 0.5)

    with pytest.raises(ValueError):
        FastFourierTransform(irregular_polar_grid, 2, 10.5)

def test_mft_precomputations():
    input_grid = make_pupil_grid(128)
    output_grid = make_fft_grid(input_grid, 1, 0.25)

    for precompute_matrices in [True, False]:
        for allocate_intermediate in [True, False]:
            mft = MatrixFourierTransform(
                input_grid, output_grid,
                precompute_matrices=precompute_matrices, allocate_intermediate=allocate_intermediate
            )

            mft.forward(input_grid.zeros())
            mft.forward(input_grid.ones())

            assert (mft.M1 is not None) == precompute_matrices
            assert (mft.intermediate_array is not None) == allocate_intermediate

def test_fourier_filter():
    for n in [16, 17, [16, 17]]:
        for q in [1, 2, 3]:
            for tensor_shape in [(), (3,), (3, 3)]:
                for scale in [1, 2]:
                    input_grid = make_pupil_grid(n, scale)

                    fft = FastFourierTransform(input_grid, q)

                    tf_shape = tensor_shape + (fft.output_grid.size,)
                    transfer_function = np.random.randn(*tf_shape) + 1j * np.random.randn(*tf_shape)
                    transfer_function = Field(transfer_function, fft.output_grid)

                    fourier_filter = FourierFilter(input_grid, transfer_function, q)

                    f_shape = tensor_shape + (input_grid.size,)
                    if len(tensor_shape) == 2:
                        f_shape = f_shape[1:]
                    f_in = Field(np.random.randn(*f_shape) + 1j * np.random.randn(*f_shape), input_grid)

                    f_out_ff = fourier_filter.forward(f_in)
                    f_in_ff = fourier_filter.backward(f_out_ff)

                    if len(tensor_shape) == 2:
                        ft = fft.forward(f_in)
                        ft = field_dot(transfer_function, ft)
                        f_out_fft = fft.backward(ft)

                        ft = fft.forward(f_out_fft)
                        ft = field_dot(field_conjugate_transpose(transfer_function), ft)
                        f_in_fft = fft.backward(ft)
                    else:
                        f_out_fft = fft.backward(fft.forward(f_in) * transfer_function)
                        f_in_fft = fft.backward(fft.forward(f_out_fft) * transfer_function.conj())

                    assert np.allclose(f_out_fft, f_out_ff)
                    assert np.allclose(f_in_fft, f_in_ff)

def check_czt_vs_scipy(x, m, w, a, dtype):
    # Check that the CZT gives the same answer as the scipy implementation.
    n = len(x)

    czt_hcipy = ChirpZTransform(n, m, w, a)
    czt_scipy = scipy.signal.CZT(n, m, w, a)

    y_hcipy = czt_hcipy(x)
    y_scipy = czt_scipy(x)

    assert y_hcipy.dtype == dtype

    rtol = 1e-13 if dtype == 'complex128' else 1e-4

    assert np.allclose(y_hcipy, y_scipy, rtol=rtol)

@pytest.mark.parametrize('dtype', ['complex128', 'complex64'])
@pytest.mark.skipif(
    version.parse(scipy.__version__) < version.parse('1.8.0'),
    reason="Requires scipy 1.8.0 or newer"
)
def test_chirp_z_transform(dtype):
    # Fix randomness.
    np.random.seed(0)

    ns = np.random.exponential(1000, size=10).astype('int')
    ms = np.random.exponential(1000, size=10).astype('int')

    for n, m in zip(ns, ms):
        x = np.random.randn(n) + 1j * np.random.randn(n)
        x = x.astype(dtype)

        w = np.exp(1j * np.random.uniform(0, 2 * np.pi))
        a = np.exp(1j * np.random.uniform(0, 2 * np.pi))

        check_czt_vs_scipy(x, m, w, a, dtype)



import numpy as np

from .multi_scale import MultiScaleCoronagraph
from ..field import Field

class FQPMCoronagraph(MultiScaleCoronagraph):
    def __init__(self, input_grid, lyot_stop=None, q=1024, scaling_factor=4, window_size=32):
        complex_mask = lambda grid: Field(np.sign(grid.x) * np.sign(grid.y), grid).astype('complex')
        super().__init__(input_grid, complex_mask, lyot_stop, q, scaling_factor, window_size)



import numpy as np
import matplotlib as mpl

def errorfill(x, y, y_err_pos, y_err_neg=None, color=None, alpha_fill=0.25, ax=None, **kwargs):
    """
    Plot a line with filled errorbars.
    """
    import matplotlib.pyplot as plt

    if ax is None:
        ax = plt.gca()

    if y_err_neg is None:
        y_err_neg = y_err_pos

    y_min = np.array(y) - np.array(y_err_neg)
    y_max = np.array(y) + np.array(y_err_pos)
    if color is None:
        l = ax.plot(x, y, **kwargs)
        color = l[0].get_color()
    else:
        l = ax.plot(x, y, color=color, **kwargs)

    facecolor = mpl.colors.colorConverter.to_rgba(color, alpha=alpha_fill)
    edgecolor = (0, 0, 0, 0)
    ax.fill_between(x, y_max, y_min, edgecolor=edgecolor, facecolor=facecolor)

    return l



__all__ = [
    'SpectralNoiseFactory',
    'SpectralNoise',
    'SpectralNoiseFactoryFFT',
    'SpectralNoiseFFT',
    'SpectralNoiseFactoryMultiscale',
    'SpectralNoiseMultiscale',
    'large_poisson',
    'large_gamma',
    'make_emccd_noise',
    'inverse_truncated',
    'inverse_truncated_modal',
    'inverse_tikhonov',
    'SVD',
    'read_fits',
    'write_fits',
    'read_grid',
    'write_grid',
    'read_field',
    'write_field',
    'read_mode_basis',
    'write_mode_basis',
    'generate_convolution_matrix',
    'make_laplacian_matrix',
    'make_derivative_matrix'
]

from .spectral_noise import *
from .stats import *
from .matrix_inversion import *
from .singular_value_decomposition import *
from .io import *
from .finite_difference import *



from .surface_profiles import parabolic_surface_sag
from .apodization import SurfaceApodizer

class ThinLens(SurfaceApodizer):
    '''A parabolic thin lens.

    Parameters
    ----------
    focal_length : scalar
        The focal length of the thin lens at the refernce wavelength.
    refractive_index : scalar or function of wavelength
        The refractive index of the lens material.
    reference_wavelength : scalar
        The wavelength for which the focal length is defined.

    '''
    def __init__(self, focal_length, refractive_index, reference_wavelength):
        self._focal_length = focal_length
        self._refractive_index = refractive_index
        self._reference_wavelength = reference_wavelength

        n0 = refractive_index(reference_wavelength)
        radius_of_curvature = focal_length * (n0 - 1)
        sag = parabolic_surface_sag(-radius_of_curvature)
        super().__init__(sag, refractive_index)

    @property
    def focal_length(self):
        '''The focal length of the lens.
        '''
        return self._focal_length

    @focal_length.setter
    def focal_length(self, focal_length):
        self._focal_length = focal_length

        n0 = self._refractive_index(self._reference_wavelength)
        radius_of_curvature = focal_length * (n0 - 1)
        sag = parabolic_surface_sag(-radius_of_curvature)

        self.surface_sag = sag



from .wavefront_sensor import WavefrontSensorOptics, WavefrontSensorEstimator
from ..optics import OpticalSystem, MicroLensArray
from ..field import CartesianGrid, Field, SeparatedCoords
from ..propagation import FresnelPropagator

import numpy as np
from scipy import ndimage

class ShackHartmannWavefrontSensorOptics(WavefrontSensorOptics):
    def __init__(self, input_grid, micro_lens_array):
        # Make propagator
        sh_prop = FresnelPropagator(input_grid, micro_lens_array.focal_length)

        # Make optical system
        OpticalSystem.__init__(self, (micro_lens_array, sh_prop))
        self.mla_index = micro_lens_array.mla_index
        self.mla_grid = micro_lens_array.mla_grid
        self.micro_lens_array = micro_lens_array

class SquareShackHartmannWavefrontSensorOptics(ShackHartmannWavefrontSensorOptics):
    ## Helper class to create a Shack-Hartmann WFS with square microlens array
    def __init__(self, input_grid, f_number, num_lenslets, pupil_diameter):
        lenslet_diameter = float(pupil_diameter) / num_lenslets
        x = np.arange(-pupil_diameter, pupil_diameter, lenslet_diameter)
        self.mla_grid = CartesianGrid(SeparatedCoords((x, x)))

        focal_length = f_number * lenslet_diameter
        self.micro_lens_array = MicroLensArray(input_grid, self.mla_grid, focal_length)

        ShackHartmannWavefrontSensorOptics.__init__(self, input_grid, self.micro_lens_array)

class ShackHartmannWavefrontSensorEstimator(WavefrontSensorEstimator):
    def __init__(self, mla_grid, mla_index, estimation_subapertures=None):
        self.mla_grid = mla_grid
        self.mla_index = mla_index
        if estimation_subapertures is None:
            self.estimation_subapertures = np.unique(self.mla_index)
        else:
            self.estimation_subapertures = np.flatnonzero(np.array(estimation_subapertures))
        self.estimation_grid = self.mla_grid.subset(estimation_subapertures)

    def estimate(self, images):
        image = images[0]

        fluxes = ndimage.measurements.sum(image, self.mla_index, self.estimation_subapertures)
        sum_x = ndimage.measurements.sum(image * image.grid.x, self.mla_index, self.estimation_subapertures)
        sum_y = ndimage.measurements.sum(image * image.grid.y, self.mla_index, self.estimation_subapertures)

        centroid_x = sum_x / fluxes
        centroid_y = sum_y / fluxes

        centroids = np.array((centroid_x, centroid_y)) - np.array(self.mla_grid.points[self.estimation_subapertures, :]).T
        return Field(centroids, self.estimation_grid)



import numpy as np
from math import pi

def binned_profile(y, x, bins=20, statistic='mean'):
    '''Create a profile of y(x) in the specified bins.

    Parameters
    ----------
    y : array_like
        The y-coordinates of the data points. This should be 1-dimensional.
    x : array_like
        The x-coordinates of the data points. This should be 1-dimensional
        and the same size as `y`.
    bins : array_like or integer
        The bin edges of the profile. If this is an integer, `bins` is the
        number of bins that will be equally distributed along the whole range
        of `x`.
    statistic : string
        The statistic to compute (default is 'mean').
        The following statistics are available:
        * 'mean': compute the mean of values for points within the bin edges while ignoring NaN values.
        * 'sum': compute the sum of values for points within the bin edges while ignoring NaN values.
        * 'min': compute the minimum of values for points within the bin edges while ignoring NaN values.
        * 'max': compute the maximum of values for points within the bin edges while ignoring NaN values.
        * 'std': compute the standard deviation of values for points within the bin edges while ignoring NaN values.
        * 'var': compute the variance of values for points within the bin edges while ignoring NaN values.

    Returns
    -------
    bin_centers : array_like
        The center of each of the bins.
    profile : array_like
        The y-values of the resulting profile.
    std_profile : array_like
        The standard deviation within each bin.
    num_per_bin : array_like
        The number of samples per bin.

    Raises
    ------
    RuntimeError
        If the number of bins are negative or zero.
    ValueError
        If the statistic is not implemented.
    '''
    if np.isscalar(bins):
        if bins <= 0:
            raise RuntimeError('The number of bins should be positive.')

        # Equally space the bins.
        bins = np.linspace(np.min(x), np.max(x), bins + 1)

    bin_centers = (bins[1:] + bins[:-1]) / 2
    num_bins = len(bin_centers)

    num_per_bin = np.histogram(x, bins)[0]
    which_bin = np.digitize(x, bins)

    statistics = {'mean': np.nanmean, 'median': np.nanmedian, 'max': np.nanmax, 'min': np.nanmin, 'std': np.nanstd, 'var': np.nanvar}

    if statistic not in statistics:
        raise ValueError('Statistic %s not implemented.' % statistic)

    func = statistics[statistic]

    profile = np.array([func(y[which_bin == b]) for b in range(1, num_bins + 1)])
    std_profile = np.array([np.nanstd(y[which_bin == b]) for b in range(1, num_bins + 1)])

    return bin_centers, profile, std_profile, num_per_bin

def azimutal_profile(image, num_bins, statistic='mean'):
    '''Create an azimuthal profile of the image around its center.

    Parameters
    ----------
    image : Field
        The image that we want an azimuthal profile from. This image must be
        two-dimensional.
    num_bins : integer
        The number of bins in theta. Bins will be equally distributed in theta.
    statistic : string
        The statistic to compute (default is 'mean').
        The following statistics are available:
        * 'mean': compute the mean of values for points within the bin edges.
        * 'sum': compute the sum of values for points within the bin edges.
        * 'min': compute the minimum of values for points within the bin edges.
        * 'max': compute the maximum of values for point within the bin edges.

    Returns
    -------
    bin_centers : array_like
        The center of each of the bins.
    profile : array_like
        The y-values of the resulting azimuthal profile.
    std_profile : array_like
        The standard deviation within each bin.
    num_per_bin : array_like
        The number of samples per bin.

    Raises
    ------
    RuntimeError
        If the number of bins are negative or zero.
    ValueError
        If the statistic is not implemented.
    '''
    theta = image.grid.as_('polar').theta
    bins = np.linspace(-pi, pi, num_bins + 1)

    return binned_profile(image, theta, bins, statistic=statistic)

def radial_profile(image, bin_size, statistic='mean'):
    '''Create a radial profile of the image around its center.

    Parameters
    ----------
    image : Field
        The image that we want a radial profile from. This image must be
        two-dimensional.
    bin_size : scalar
        The extent of each bin. Each bin will be a ring from r to r+`bin_size`.
    statistic : string
        The statistic to compute (default is 'mean').
        The following statistics are available:
        * 'mean': compute the mean of values for points within the bin edges.
        * 'sum': compute the sum of values for points within the bin edges.
        * 'min': compute the minimum of values for points within the bin edges.
        * 'max': compute the maximum of values for point within the bin edges.

    Returns
    -------
    bin_centers : array_like
        The center of each of the bins.
    profile : array_like
        The y-values of the resulting radial profile.
    std_profile : array_like
        The standard deviation within each bin.
    num_per_bin : array_like
        The number of samples per bin.

    Raises
    ------
    ValueError
        If the statistic is not implemented.
    '''
    r = image.grid.as_('polar').r

    n_bins = int(np.ceil(r.max() / bin_size))
    max_bin = n_bins * bin_size
    bins = np.linspace(0, max_bin, n_bins + 1)

    return binned_profile(image, r, bins, statistic=statistic)



import numpy as np
from ..field import Field
from .wavefront import Wavefront

class GaussianBeam(object):
    '''An analytical description of a light beam with a Gaussian profile.

    Parameters
    ----------
    w0 : scalar
        The beam waist.
    z : scalar
        The current distance from the beam waist.
    wavelength : scalar
        The wavelength of the light.
    '''
    def __init__(self, w0, z, wavelength):
        self.w0 = w0
        self.z = z
        self.wavelength = wavelength

    @property
    def beam_waist(self):
        '''The beam waist of the Gaussian beam.
        '''
        return self.w0

    @beam_waist.setter
    def beam_waist(self, w0):
        self.w0 = w0

    @property
    def zR(self):  # noqa: N802
        '''The Rayleight distance of the Gaussian beam.
        '''
        return np.pi * self.w0**2 / self.wavelength

    @zR.setter
    def zR(self, zR):  # noqa: N802
        self.w0 = np.sqrt(zR * self.wavelength / np.pi)

    rayleigh_distance = zR

    @property
    def q(self):
        '''The complex beam parameter of the Gaussian beam.
        '''
        return self.z + 1j * self.zR

    @q.setter
    def q(self, q):
        self.z = np.real(q)
        self.zR = np.imag(q)

    complex_beam_parameter = q

    @property
    def theta(self):
        '''The beam divergence of the Gaussian beam.
        '''
        return self.wavelength / (np.pi * self.w0)

    @theta.setter
    def theta(self, theta):
        self.w0 = self.wavelength / (theta * np.pi)

    beam_divergence = theta

    @property
    def R(self):  # noqa: N802
        '''The current radius of curvature of the Gaussian beam.
        '''
        epsilon = 1e-16
        if abs(self.z) < epsilon:
            return np.inf
        else:
            return self.z * (1 + (self.zR / self.z)**2)

    radius_of_curvature = R

    @property
    def psi(self):
        '''The current Gouy phase of the Gaussian beam.
        '''
        return np.arctan(self.z / self.zR)

    gouy_phase = psi

    @property
    def w(self):
        '''The current beam radius of the Gaussian beam.
        '''
        return self.w0 * np.sqrt(1 + (self.z / self.zR)**2)

    beam_radius = w

    @property
    def FWHM(self):  # noqa: N802
        '''The current FWHM of the Gaussian beam.
        '''
        return self.w * np.sqrt(2 * np.log(2))

    full_width_half_maximum = FWHM

    @property
    def k(self):
        '''The wavenumber of the Gaussian beam.
        '''
        return 2 * np.pi / self.wavelength

    @k.setter
    def k(self, k):
        self.wavelength = 2 * np.pi / k

    wavenumber = k

    def evaluate(self, grid):
        '''Evaluate the wavefront of the Gaussian beam at the current position on
        the given grid.

        Parameters
        ----------
        grid : Grid
            The grid on which to calculate the wavefront for the Gaussian beam.

        Returns
        -------
        Wavefront
            The evaluated wavefront of the Gaussian beam.
        '''
        if grid.is_('cartesian'):
            r2 = grid.x**2 + grid.y**2
        else:
            r2 = grid.as_('polar').r**2

        K1 = self.w0 / self.w
        K2 = np.exp(-r2 / self.w**2)
        K3 = np.exp(-1j * (self.k * self.z + self.k * r2 / (2 * self.R) - self.psi))

        return Wavefront(Field(K1 * K2 * K3, grid), self.wavelength)

    __call__ = evaluate



import numpy as np
from .optical_element import OpticalElement
from ..field import CartesianGrid, UnstructuredCoords

class PeriodicOpticalElement(OpticalElement):
    def __init__(self, input_grid, pitch, apodization, orientation=0, even_grid=False):
        '''An even asphere micro-lens array.

        Parameters
        ----------
        input_grid : Grid
            The grid on which the periodic optical element is evaluated.
        pitch : scalar
            The pitch of the periodic optical element.
        apodization : Apodizer
            The apodizer that will be evaluated on the periodic grid.
        orientation : scalar
            The orientation of the periodic optical element.
        even_grid : bool
            This determines whether zero is in between two elements or if it is the center of an element.
        '''
        self.input_grid = input_grid.copy()
        self.input_grid = self.input_grid.rotated(orientation)

        if even_grid:
            xf = (np.fmod(abs(self.input_grid.x), pitch) - pitch / 2) * np.sign(self.input_grid.x)
            yf = (np.fmod(abs(self.input_grid.y), pitch) - pitch / 2) * np.sign(self.input_grid.y)
        else:
            xf = (np.fmod(abs(self.input_grid.x) + pitch / 2, pitch) - pitch / 2) * np.sign(self.input_grid.x)
            yf = (np.fmod(abs(self.input_grid.y) + pitch / 2, pitch) - pitch / 2) * np.sign(self.input_grid.y)

        periodic_grid = CartesianGrid(UnstructuredCoords((xf, yf)))
        self.apodization = apodization(periodic_grid)

    def forward(self, wavefront):
        wf = wavefront.copy()
        return self.apodization.forward(wf)

    def backward(self, wavefront):
        wf = wavefront.copy()
        return self.apodization.backward(wf)



from hcipy import *
import numpy as np
import os
import warnings
import pytest
import functools

def check_aperture(aperture_function, diameter, name, check_normalization=False, check_segmentation=False, **aperture_args):
    fname = os.path.join(os.path.dirname(__file__), 'baseline_for_apertures/' + name + '.fits.gz')

    grid = make_uniform_grid(256, [diameter, diameter])
    field = evaluate_supersampled(aperture_function(**aperture_args), grid, 8)

    if check_normalization:
        grid_norm = grid.scaled(1 / diameter)
        field_norm = evaluate_supersampled(aperture_function(normalized=True, **aperture_args), grid_norm, 8)

        assert np.allclose(field, field_norm)

    if check_segmentation:
        aperture, segments = aperture_function(**aperture_args, return_segments=True)

        aperture = evaluate_supersampled(aperture, grid, 1)
        segments = evaluate_supersampled(segments, grid, 1)

        aperture_from_segments = segments.linear_combination(np.ones(len(segments)))

        assert np.allclose(aperture, aperture_from_segments)

    if os.path.isfile(fname):
        reference = read_fits(fname).ravel()

        assert np.allclose(field, reference)
    else:
        warnings.warn('Baseline aperture %s not available. Writing reference now...' % fname)

        if not os.path.exists(os.path.dirname(fname)):
            os.makedirs(os.path.dirname(fname))
        write_fits(field, fname)

def test_regular_polygon_aperture():
    pupil_diameter = 1

    for num_sides in [6, 5]:
        for angle in [0, 15]:
            name = 'polygon/pupil'
            name += '_rotated' if angle != 0 else ''
            name += '_hex' if num_sides == 6 else '_pent'

            check_aperture(
                make_regular_polygon_aperture, pupil_diameter, name,
                circum_diameter=pupil_diameter, num_sides=num_sides, angle=angle
            )

def test_circular_aperture():
    pupil_diameter = 1

    for diameter in [1, 0.5]:
        name = 'circular/pupil'
        name += '_small' if diameter < 0.7 else '_large'

        # Use a functools.partial() here to avoid the name collision of the diameter argument.
        check_aperture(functools.partial(make_circular_aperture, diameter=diameter), pupil_diameter, name)

def test_rectangular_aperture():
    pupil_diameter = 1

    for size in [[0.5, 0.5], [1, 0.5]]:
        name = 'rectangular/pupil'
        name += '_square' if size[0] == size[1] else '_elongated'

        check_aperture(make_rectangular_aperture, pupil_diameter, name, size=size)

def test_elliptical_aperture():
    pupil_diameter = 1

    for diameters in [[0.5, 0.5], [1, 0.5]]:
        name = 'ellipse/pupil'
        name += '_round' if diameters[0] == diameters[1] else '_elongated'

        check_aperture(make_elliptical_aperture, pupil_diameter, name, diameters=diameters)

def test_obstructed_circular_aperture():
    pupil_diameter = 1

    for central_obscuration_ratio in [0.1, 0.3]:
        for num_spiders in [3, 5]:
            for spider_width in [0.01, 0.03]:
                name = 'obstructed_circular/pupil'
                name += '_small_obscuration' if central_obscuration_ratio < 0.2 else '_large_obscuration'
                name += f'_{num_spiders}spiders'
                name += '_thinspiders' if spider_width < 0.02 else '_thickspiders'

                check_aperture(
                    make_obstructed_circular_aperture, pupil_diameter, name,
                    pupil_diameter=pupil_diameter, central_obscuration_ratio=central_obscuration_ratio, num_spiders=num_spiders, spider_width=spider_width
                )

@pytest.mark.parametrize('num_rings', [3, 5])
@pytest.mark.parametrize('segment_flat_to_flat', [0.04, 0.07])
@pytest.mark.parametrize('gap_size', [0.01, 0.02])
@pytest.mark.parametrize('starting_ring', [0, 2])
def test_hexagonal_segmented_aperture(num_rings, segment_flat_to_flat, gap_size, starting_ring):
    name = 'hexagonal_segmented/pupil'
    name += '_smallgap' if gap_size < 0.015 else '_largegap'
    name += f'_{num_rings}rings'
    name += '_smallsegment' if segment_flat_to_flat < 0.05 else '_largesegment'
    name += '_withcenter' if starting_ring == 0 else '_withoutcenter'

    check_aperture(
        make_hexagonal_segmented_aperture, 1, name,
        check_segmentation=True,
        num_rings=num_rings, segment_flat_to_flat=segment_flat_to_flat, gap_size=gap_size, starting_ring=starting_ring
    )

@pytest.mark.parametrize('telescope', ['ut1', 'ut2', 'ut3', 'antu', 'kueyen', 'melipal'])
@pytest.mark.parametrize('with_spiders', [True, False])
def test_vlt_ut_123_aperture(telescope, with_spiders):
    name = 'vlt/pupil_ut123'
    name += '_without_spiders' if not with_spiders else ''

    check_aperture(
        make_vlt_aperture, 8.0, name,
        check_normalization=True,
        check_segmentation=True,
        telescope=telescope, with_spiders=with_spiders
    )

@pytest.mark.parametrize('telescope', ['ut4', 'yepun'])
@pytest.mark.parametrize('with_spiders', [True, False])
@pytest.mark.parametrize('with_M3_cover', [True, False])
def test_vlt_ut_4_aperture(telescope, with_spiders, with_M3_cover):
    name = 'vlt/pupil_ut4'
    name += '_without_spiders' if not with_spiders else ''
    name += '_with_M3_cover' if with_M3_cover else ''

    check_aperture(
        make_vlt_aperture, 8.1196, name,
        check_normalization=True,
        check_segmentation=True,
        telescope=telescope, with_spiders=with_spiders, with_M3_cover=with_M3_cover
    )

def test_invalid_vlt_apertures():
    for telescope in ['ut1', 'ut2', 'ut3']:
        with pytest.warns(UserWarning, match='Using the M3 cover on a telescope other than UT4 is not realistic.'):
            make_vlt_aperture(telescope=telescope, with_M3_cover=True)

    with pytest.raises(ValueError):
        make_vlt_aperture(telescope='nonexistent_vlt_telescope')


@pytest.mark.parametrize('with_spiders', [True, False])
def test_vlti_aperture(with_spiders):
    name = 'vlti/pupil'
    name += '_without_spiders' if not with_spiders else ''

    check_aperture(
        make_vlti_aperture, 125.0, name,
        check_normalization=False, check_segmentation=True,
        with_spiders=with_spiders, zenith_angle=0.0, azimuth=0.0
    )

    name = 'vlti/pupil_non_zenith'
    name += '_without_spiders' if not with_spiders else ''

    check_aperture(
        make_vlti_aperture, 125.0, name,
        check_normalization=False, check_segmentation=True,
        with_spiders=with_spiders, zenith_angle=np.pi / 4, azimuth=np.pi / 4
    )

@pytest.mark.parametrize('with_spiders', [True, False])
def test_magellan_aperture(with_spiders):
    name = 'magellan/pupil'
    name += '_without_spiders' if not with_spiders else ''

    check_aperture(
        make_magellan_aperture, 6.5, name,
        check_normalization=True,
        with_spiders=with_spiders
    )

@pytest.mark.parametrize('with_spiders', [True, False])
def test_hale_aperture(with_spiders):
    name = 'hale/pupil'
    name += '_without_spiders' if not with_spiders else ''

    check_aperture(
        make_hale_aperture, 5.08, name,
        check_normalization=True,
        with_spiders=with_spiders
    )

@pytest.mark.parametrize('with_spiders', [True, False])
@pytest.mark.parametrize('with_segment_gaps', [True, False])
def test_luvoir_a_aperture(with_spiders, with_segment_gaps):
    name = 'luvoir_a/pupil'
    name += '_without_spiders' if not with_spiders else ''
    name += '_without_segment_gaps' if not with_segment_gaps else ''

    check_aperture(
        make_luvoir_a_aperture, 15, name,
        check_normalization=True, check_segmentation=True,
        with_spiders=with_spiders, with_segment_gaps=with_segment_gaps
    )

def test_luvoir_a_lyot_stop():
    for with_spiders in [True, False]:
        for spider_oversize in [1, 3]:
            for outer_diameter_fraction in [0.8, 0.9]:
                name = 'luvoir_a_lyot/pupil'
                name += '_withoutspiders' if not with_spiders else ''
                name += f'_{spider_oversize}spideroversize'
                name += f'_{int(outer_diameter_fraction * 100)}od'

                check_aperture(
                    make_luvoir_a_lyot_stop, 15, name,
                    check_normalization=True,
                    with_spiders=with_spiders, spider_oversize=spider_oversize, outer_diameter_fraction=outer_diameter_fraction
                )

@pytest.mark.parametrize('with_segment_gaps', [True, False])
def test_luvoir_b_aperture(with_segment_gaps):
    name = 'luvoir_b/pupil'
    name += '_without_segment_gaps' if not with_segment_gaps else ''

    check_aperture(
        make_luvoir_b_aperture, 8, name,
        check_normalization=True, check_segmentation=True,
        with_segment_gaps=with_segment_gaps
    )

@pytest.mark.parametrize('with_spiders', [True, False])
@pytest.mark.parametrize('with_segment_gaps', [True, False])
def test_hicat_aperture(with_spiders, with_segment_gaps):
    name = 'hicat_pupil/pupil'
    name += '_without_spiders' if not with_spiders else ''
    name += '_without_segment_gaps' if not with_segment_gaps else ''

    check_aperture(
        make_hicat_aperture, 0.019725, name,
        check_normalization=True, check_segmentation=True,
        with_spiders=with_spiders, with_segment_gaps=with_segment_gaps
    )

@pytest.mark.parametrize('with_spiders', [True, False])
def test_hicat_lyot(with_spiders):
    name = 'hicat_lyot/pupil'
    name += '_without_spiders' if not with_spiders else ''

    check_aperture(
        make_hicat_lyot_stop, 0.01915751488095238, name,
        check_normalization=True,
        with_spiders=with_spiders
    )

@pytest.mark.parametrize('with_spiders', [True, pytest.param(False, marks=pytest.mark.slow)])
def test_elt_aperture(with_spiders):
    name = 'elt/pupil'
    name += '_without_spiders' if not with_spiders else ''

    check_aperture(
        make_elt_aperture, 39.14634, name,
        check_normalization=True, check_segmentation=True,
        with_spiders=with_spiders
    )

@pytest.mark.parametrize('with_spiders', [True, pytest.param(False, marks=pytest.mark.slow)])
def test_tmt_aperture(with_spiders):
    name = 'tmt/pupil'
    name += '_without_spiders' if not with_spiders else ''

    check_aperture(
        make_tmt_aperture, 30.0, name,
        check_normalization=True, check_segmentation=True,
        with_spiders=with_spiders
    )

@pytest.mark.parametrize('with_spiders', [True, False])
def test_gmt_aperture(with_spiders):
    name = 'gmt/pupil'
    name += '_without_spiders' if not with_spiders else ''

    check_aperture(
        make_gmt_aperture, 25.448, name,
        check_normalization=True, check_segmentation=True,
        with_spiders=with_spiders
    )

def test_habex_aperture():
    name = 'habex/pupil'

    check_aperture(make_habex_aperture, 4.0, name)

@pytest.mark.parametrize('with_spiders', [True, False])
@pytest.mark.parametrize('with_pads', [True, False])
def test_hst_aperture(with_spiders, with_pads):
    name = 'hst/pupil'
    name += '_without_spiders' if not with_spiders else ''
    name += '_without_pads' if not with_pads else ''

    check_aperture(
        make_hst_aperture, 2.4, name,
        check_normalization=True,
        with_spiders=with_spiders, with_pads=with_pads
    )

@pytest.mark.parametrize('with_spiders', [True, False])
def test_jwst_aperture(with_spiders):
    name = 'jwst/pupil'
    name += '_without_spiders' if not with_spiders else ''

    check_aperture(
        make_jwst_aperture, 6.603464, name,
        check_normalization=True, check_segmentation=True,
        with_spiders=with_spiders
    )

def test_shifted_aperture():
    grid = make_pupil_grid(256, 2.0)
    aperture1 = make_circular_aperture(1.0, center=[0.25, 0.25])(grid)
    aperture2 = make_shifted_aperture(make_circular_aperture(1.0), np.array([0.25, 0.25]))(grid)

    assert np.allclose(aperture1, aperture2)

@pytest.mark.parametrize('with_spiders', [True, False])
@pytest.mark.parametrize('with_segment_gaps', [True, False])
def test_keck_aperture(with_spiders, with_segment_gaps):
    name = 'keck/pupil'
    name += '_without_spiders' if not with_spiders else ''
    name += '_without_segment_gaps' if not with_segment_gaps else ''

    check_aperture(
        make_keck_aperture, 10.95, name,
        check_normalization=True, check_segmentation=True,
        with_spiders=with_spiders, with_segment_gaps=with_segment_gaps
    )



from __future__ import division

import numpy as np
from .coordinates import RegularCoords, SeparatedCoords, UnstructuredCoords
from .field import Field
from .cartesian_grid import CartesianGrid
from .grid import Grid

def make_uniform_grid(dims, extent, center=0, has_center=False):
    '''Create a uniformly-spaced :class:`Grid` of a certain shape and size.

    Parameters
    ----------
    dims : scalar or ndarray
        The number of points in each dimension. If this is a scalar, it will
        be multiplexed over all dimensions.
    extent : scalar or ndarray
        The total extent of the grid in each dimension.
    center : scalar or ndarray
        The center point. The grid will by symmetric around this point.
    has_center : boolean
        Does the grid has to have the center as one of its points. If this is
        False, this does not mean that the grid will not have the center.

    Returns
    -------
    Grid
        A :class:`Grid` with :class:`RegularCoords`.
    '''
    num_dims = max(np.array([dims]).shape[-1], np.array([extent]).shape[-1], np.array([center]).shape[-1])

    dims = (np.ones(num_dims) * dims).astype('int')
    extent = (np.ones(num_dims) * extent).astype('float')
    center = (np.ones(num_dims) * center).astype('float')

    delta = extent / dims
    zero = -extent / 2 + center + delta / 2

    if has_center:
        zero -= delta / 2 * (1 - np.mod(dims, 2))

    return CartesianGrid(RegularCoords(delta, dims, zero))

def make_pupil_grid(dims, diameter=1):
    '''Makes a new :class:`Grid`, meant for descretisation of a pupil-plane wavefront.

    This grid is symmetric around the origin, and therefore has no point exactly on
    the origin for an even number of pixels.

    Parameters
    ----------
    dims : ndarray or integer
        The number of pixels per dimension. If this is an integer, this number
        of pixels is used for all dimensions.
    diameter : ndarray or scalar
        The diameter of the grid in each dimension. If this is a scalar, this diameter
        is used for all dimensions.

    Returns
    -------
    Grid
        A :class:`CartesianGrid` with :class:`RegularCoords`.
    '''
    diameter = (np.ones(2) * diameter).astype('float')
    return make_uniform_grid(dims, diameter)

def make_focal_grid_from_pupil_grid(pupil_grid, q=1, num_airy=None, focal_length=1, wavelength=1):
    '''Make a grid for a focal plane from a pupil grid.

    Calculate the focal grid corresponding to the pupil grid, using an FFT grid as a guide. The
    resulting grid focal will always contain the origin (0, 0) point. The extent of the pupil
    grid will be used as the diameter of the pupil. If the pupil is undersized on the pupil
    grid, the resulting focal grid needs to be rescaled manually.

    .. note::
        In almost all cases, it is preferable to use :func:`make_focal_grid`. This
        function allows you to directly set the diameter, and doesn't require the user to pass
        the pupil grid as an argument. `make_focal_grid_from_pupil_grid()` retains old functionality
        and serves as a backwards compatibility function, due to its ubiquitous usage in HCIPy code.

    Parameters
    ----------
    pupil_grid : Grid
        The pupil grid for which the focal grid needs to be calculated. The extent of this Grid
        is used as the diameter of the pupil.
    q : scalar or array_like
        The number of pixels per resolution element (= lambda f / D).
    num_airy : scalar or array_like
        The spatial extent of the grid in radius in resolution elements (= lambda f / D).
    focal_length : scalar
        The focal length used for calculating the spatial resolution at the focal plane.
    wavelength : scalar
        The reference wavelength used for calculating the spatial resolution at the focal plane.

    Returns
    -------
    Grid
        A Grid describing the sampling for a focal plane.
    '''
    from ..fourier import make_fft_grid

    f_lambda = focal_length * wavelength
    if num_airy is None:
        fov = 1
    else:
        fov = (num_airy * np.ones(pupil_grid.ndim, dtype='float')) / (pupil_grid.shape / 2)

    if np.max(fov) > 1:
        import warnings
        warnings.warn('Focal grid is larger than the maximum allowed angle (fov=%.03f). You may see wrapping when doing propagations.' % np.max(fov), stacklevel=2)

    uv = make_fft_grid(pupil_grid, q, fov)
    focal_grid = uv.scaled(f_lambda / (2 * np.pi))

    return focal_grid

def make_focal_grid(q, num_airy, spatial_resolution=None, pupil_diameter=None, focal_length=None, f_number=None, reference_wavelength=None):
    r'''Make a grid for a focal plane.

    This grid will be a CartesianGrid with RegularCoords, and supports different resolutions, samplings,
    or extent in x and y. If `spatial_resolution` is 1, then the grid will be returned in normalized
    (ie. homogenized) coordinates. Otherwise, it will be in physical units. The spatial resolution is
    defined by:

    .. math:: \Delta x = \lambda f / D = \lambda F

    where :math:`\lambda` is the wavelength, :math:`f` is the effective focal length before the focal plane,
    :math:`D` is the diameter of the pupil, and :math:`F` is the F-number of the incoming light beam.
    You can supply either the spatial resolution or a set of focal length, reference wavelength and pupil diameter,
    or a set of F-number and reference wavelength. If none are supplied, a spatial resolution of 1 will be
    assumed, meaning normalized units.

    The grid will always contain the origin (0, 0) point.

    Parameters
    ----------
    q : scalar or array_like
        The number of pixels per resolution element (= lambda f / D).
    num_airy : scalar or array_like
        The spatial extent of the grid in radius in resolution elements (= lambda f / D).
    spatial_resolution : scalar    or array_like
        The physical size of a resolution element (= lambda f / D). It this is not given,
        the spatial resolution will be calculated from the given `focal_length`,
        `reference_wavelength` and `pupil_diameter`.
    pupil_diameter : scalar or array_like
        The diameter of the pupil. If it is an array, this indicates the diameter in x and y.
    focal_length : scalar
        The focal length used for calculating the spatial resolution at the focal plane.
    f_number : scalar or array_like
        The F number, also known as focal ratio, at the focal plane. If this is given, it overrides
        the given pupil diameter and focal length.
    reference_wavelength : scalar
        The reference wavelength used for calculating the spatial resolution at the focal plane.

    Returns
    -------
    Grid
        A Grid describing the sampling for a focal plane.

    Raises
    ------
    ValueError
        If both no spatial resolution and no complete set of (focal length, reference wavelength and pupil diameter) was supplied.
    '''
    if isinstance(q, Grid):
        raise ValueError('The function signature was changed as of HCIPy 0.3.0. Please use the new signature (prefered), or use make_focal_grid_from_pupil_grid() if you want to retain old behaviour.')

    if spatial_resolution is None:
        if f_number is None:
            if pupil_diameter is None or focal_length is None:
                if reference_wavelength is None:
                    spatial_resolution = 1
                else:
                    raise ValueError('You only supplied a reference wavelength and forgot to supply either an f_number or a (pupil_diameter, focal_length).')
            else:
                f_number = focal_length / pupil_diameter

        if spatial_resolution is None:
            if reference_wavelength is None:
                raise ValueError('You supplied an f_number or (pupil_diameter, focal_length), and forgot to supply a reference wavelength.')
            else:
                spatial_resolution = f_number * reference_wavelength

    delta = spatial_resolution / q * np.ones(2)
    dims = (2 * num_airy * q * np.ones(2)).astype('int')
    zero = delta * (-dims / 2 + np.mod(dims, 2) * 0.5)

    return CartesianGrid(RegularCoords(delta, dims, zero))

def make_hexagonal_grid(circum_diameter, n_rings, pointy_top=False, center=None):
    '''Make a regular hexagonal grid.

    Parameters
    ----------
    circum_diameter : scalar
        The circum diameter of the hexagons in the grid.
    n_rings : integer
        The number of rings in the grid.
    pointy_top : boolean
        If the hexagons contained in the grid.
    center : ndarray
        The center of the grid in cartesian coordinates.

    Returns
    -------
    Grid
        A :class:`CartesianGrid` with `UnstructuredCoords`, indicating the
        center of the hexagons.
    '''
    if center is None:
        center = np.zeros(2)

    apothem = circum_diameter * np.sqrt(3) / 4

    q = [0]
    r = [0]

    for n in range(1, n_rings + 1):
        # top
        q += list(range(n, 0, -1))
        r += list(range(0, n))
        # right top
        q += list(range(0, -n, -1))
        r += [n] * n
        # right bottom
        q += [-n] * n
        r += list(range(n, 0, -1))
        # bottom
        q += list(range(-n, 0))
        r += list(range(0, -n, -1))
        # left bottom
        q += list(range(0, n))
        r += [-n] * n
        # left top
        q += [n] * n
        r += list(range(-n, 0))

    x = (-np.array(q) + np.array(r)) * circum_diameter / 2 + center[0]
    y = (np.array(q) + np.array(r)) * apothem * 2 + center[1]

    weight = 2 * apothem**2 * np.sqrt(3)

    if pointy_top:
        return CartesianGrid(UnstructuredCoords((x, y)), weight)
    else:
        return CartesianGrid(UnstructuredCoords((y, x)), weight)

def make_chebyshev_grid(dims, minimum=None, maximum=None):
    if minimum is None:
        minimum = -1

    if maximum is None:
        maximum = 1

    dims = np.array(dims)
    minimum = np.ones(len(dims)) * minimum
    maximum = np.ones(len(dims)) * maximum

    middles = (minimum + maximum) / 2
    intervals = (maximum - minimum) / 2

    sep_coords = []
    for dim, middle, interval in zip(dims, middles, intervals):
        c = np.cos(np.pi * (2 * np.arange(dim) + 1) / (2.0 * dim))
        c = middle + interval * c
        sep_coords.append(c)

    return CartesianGrid(SeparatedCoords(sep_coords))

def make_supersampled_grid(grid, oversampling):
    '''Make a new grid that oversamples by a factor `oversampling`.

    .. note ::
        The Grid `grid` must be a grid with separable coordinates.

    Parameters
    ----------
    grid : Grid
        The grid that we want to oversample.
    oversampling : integer or scalar or ndarray
        The factor by which to oversample. If this is a scalar, it will be rounded to
        the nearest integer. If this is an array, a different oversampling factor will
        be used for each dimension.

    Returns
    -------
    Grid
        The oversampled grid.
    '''
    oversampling = (np.round(oversampling)).astype('int')

    if grid.is_regular:
        delta_new = grid.delta / oversampling
        zero_new = grid.zero - grid.delta / 2 + delta_new / 2
        dims_new = grid.dims * oversampling

        return grid.__class__(RegularCoords(delta_new, dims_new, zero_new))
    elif grid.is_separated:
        raise NotImplementedError()

    raise ValueError('Cannot create a supersampled grid from a non-separated grid.')

def make_subsampled_grid(grid, undersampling):
    '''Make a new grid that undersamples by a factor `undersampling`.

    .. note ::
        The dimensions of the `grid` must be divisible by `undersampling`.

    Parameters
    ----------
    grid : Grid
        The grid that we want to oversample.
    undersampling : integer or scalar or ndarray
        The factor by which to undersample. If this is a scalar, it will be rounded to
        the nearest integer. If this is an array, a different undersampling factor will
        be used for each dimension.

    Returns
    -------
    Grid
        The undersampled grid.
    '''
    undersampling = (np.round(undersampling)).astype('int')

    if grid.is_regular:
        delta_new = grid.delta * undersampling
        zero_new = grid.zero - grid.delta / 2 + delta_new / 2
        dims_new = grid.dims // undersampling

        return grid.__class__(RegularCoords(delta_new, dims_new, zero_new))
    elif grid.is_separated:
        raise NotImplementedError()

    raise ValueError("Cannot create a subsampled grid from a non-separated grid.")

def subsample_field(field, subsampling, new_grid=None, statistic='mean'):
    '''Average the field over subsampling pixels in each dimension.

    .. note ::
        The dimensions of the grid of `field` must be divisible by `subsampling`.

    Parameters
    ----------
    field : Field
        The field to subsample. The grid of this field must have the right
        dimensions to be able to be subsampled.
    subsampling : integer or scalar or ndarray
        The subsampling factor. If this is a scalar, it will be rounded to the
        nearest integer. If this is an array, the subsampling factor will be
        different for each dimension.
    new_grid : Grid
        If this grid is given, no new grid will be calculated and this grid will
        be used instead. This saves on calculation time if your new grid is already
        known beforehand.
    statistic : string
        The statistic to compute (default is 'mean').
        The following statistics are available:
        * 'mean': compute the mean of values for points within each superpixel.
        * 'sum': compute the sum of values for points within each superpixel. This is identical to a weighted histogram.
        * 'min': compute the minimum of values for points within each superpixel.
        * 'max': compute the maximum of values for points within each superpixel.
        * 'median': compute the median of values for points within each superpixel.
        * 'nanmedian': compute the median of values for points within each superpixel while ignoring NaN values.

    Returns
    -------
    Field
        The subsampled field.
    '''
    subsampling = (np.round(subsampling)).astype('int')

    if new_grid is None:
        new_grid = make_subsampled_grid(field.grid, subsampling)

    reshape = []
    axes = []
    for i, s in enumerate(new_grid.shape):
        reshape.extend([s, subsampling])
        axes.append(2 * i + 1)

    if field.tensor_order > 0:
        reshape = list(field.tensor_shape) + reshape
        axes = np.array(axes) + field.tensor_order
        new_shape = list(field.tensor_shape) + [-1]
    else:
        new_shape = [-1]

    available_statistics = {
        'mean': np.mean,
        'max': np.max,
        'min': np.min,
        'sum': np.sum,
        'median': np.median,
        'nanmedian': np.nanmedian
    }

    if statistic not in available_statistics:
        raise ValueError('This statistic is not recognized.')

    if field.grid.is_regular:
        # All weights will be the same, so the array can be combined without taking the weights into account.
        return Field(available_statistics[statistic](field.reshape(tuple(reshape)), axis=tuple(axes)).reshape(tuple(new_shape)), new_grid)
    else:
        # Some weights will be different so calculate weighted mean instead.
        if statistic in ['min', 'max', 'sum', 'median', 'nanmedian']:
            f = available_statistics[statistic](field.reshape(tuple(reshape)), axis=tuple(axes))
            return Field(f.reshape(tuple(new_shape)), new_grid)
        else:
            # Statistic is mean
            weights = field.grid.weights
            w = weights.reshape(tuple(reshape)).sum(axis=tuple(axes))
            f = np.sum((field * weights).reshape(tuple(reshape)), axis=tuple(axes))
            return Field((f / w).reshape(tuple(new_shape)), new_grid)

def evaluate_supersampled(field_generator, grid, oversampling, statistic='mean', make_sparse=True):
    '''Evaluate a Field generator on `grid`, with an oversampling.

    Parameters
    ----------
    field_generator : Field generator or list of Field generators
        The field generator to evaluate. If this is a list of Field generators,
        each Field generator will be evaluated and stored in a ModeBasis.
    grid : Grid
        The grid on which to evaluate `field_generator`.
    oversampling : integer or scalar or ndarray
        The factor by which to oversample. If this is a scalar, it will be rounded to
        the nearest integer. If this is an array, a different oversampling factor will
        be used for each dimension.
    statistic : string
        The statistic to compute (default is 'mean').
        The following statistics are available:
        * 'mean': compute the mean of values for points within each superpixel.
        * 'sum': compute the sum of values for points within each superpixel. This is identical to a weighted histogram.
        * 'min': compute the minimum of values for points within each superpixel.
        * 'max': compute the maximum of values for point within each superpixel.
    make_sparse : boolean
        If the resulting ModeBasis needs to be sparsified. This is ignored if
        only a single Field generator is provided.

    Returns
    -------
    Field or ModeBasis
        The evaluated field or mode basis.
    '''
    import scipy.sparse
    from ..mode_basis import ModeBasis

    if isinstance(field_generator, (list, tuple)):
        modes = []

        for fg in field_generator:
            field = evaluate_supersampled(fg, grid, oversampling, statistic)

            if make_sparse:
                field = scipy.sparse.csr_matrix(field)
                field.eliminate_zeros()

            modes.append(field)

        return ModeBasis(modes, grid)

    oversampling = (np.round(oversampling) * np.ones(grid.ndim)).astype('int')

    if grid.is_separated:
        # Use dithered sub grids to get the final supersampled field.
        deltas = []
        for i in range(grid.ndim):
            x = grid.coords.separated_coords[i]
            d = (x[2:] - x[:-2]) / 2.
            d = np.concatenate(([x[1] - x[0]], d, [x[-1] - x[-2]]))
            deltas.append(d)

        dithers = make_uniform_grid(oversampling, 1)

        separated_coords = grid.separated_coords

        if statistic in ['mean', 'sum']:
            field = 0
        else:
            field = None

        for dither in dithers.points:
            dithered_separated_coords = [c + d * delta for c, d, delta in zip(separated_coords, dither, deltas)]
            dithered_grid = grid.__class__(SeparatedCoords(dithered_separated_coords))

            if statistic in ['mean', 'sum']:
                field += field_generator(dithered_grid)
            else:
                if field is None:
                    field = field_generator(dithered_grid)
                elif statistic == 'min':
                    field = np.minimum(field, field_generator(dithered_grid))
                else:
                    field = np.maximum(field, field_generator(dithered_grid))

        if statistic == 'mean':
            field /= len(dithers)

        field.grid = grid
        return field
    else:
        # Cannot use sub grids, so fall back to evaluation of generator on the full
        # supersampled grid.
        supersampled_grid = make_supersampled_grid(grid, oversampling)
        field = field_generator(supersampled_grid)
        return subsample_field(field, oversampling, grid, statistic)

def make_uniform_vector_field(field, jones_vector):
    '''Make an uniform vector field from a scalar field and a jones vector.

    Parameters
    ----------
    field : Field
        An input scalar field that is expanded to a vector field
    jones_vector : array_like
        The output vector at every pixel

    Returns
    -------
    Field
        The expanded vector field
    '''
    if field.is_scalar_field():
        return Field([ei * field for ei in jones_vector], field.grid)

def make_uniform_vector_field_generator(field_generator, jones_vector):
    '''Make an uniform vector field generator from a scalar field generator and a jones vector.

    Parameters
    ----------
    field_generator : Field generator
        The field generator to evaluate.
    jones_vector : array_like
        The output vector at every grid coordinate

    Returns
    -------
    Field generator
        This function can be evaluated on a grid to get a Field.
    '''

    def func(grid):
        scalar_field = field_generator(grid)
        return Field([ei * scalar_field for ei in jones_vector], grid)

    return func



from ..config import Configuration
from .cpu import get_num_available_cores

import functools
import numpy as np
import scipy
import warnings

_CPU_COUNT = get_num_available_cores()

try:
    import mkl_fft
except ImportError:
    mkl_fft = None

try:
    import pyfftw

    # Set default cache variables for PyFFTW.
    pyfftw.interfaces.cache.enable()
    pyfftw.interfaces.cache.set_keepalive_time(30)
except ImportError:
    pyfftw = None

def _make_func(func_name):
    if mkl_fft is not None:
        try:
            mkl_func = getattr(mkl_fft, func_name)
        except AttributeError:
            mkl_func = None

    if pyfftw is not None:
        pyfftw_func = getattr(pyfftw.interfaces.scipy_fft, func_name)

    scipy_func = getattr(scipy.fft, func_name)
    numpy_func = getattr(np.fft, func_name)

    @functools.wraps(numpy_func)
    def func(x, *args, overwrite_x=False, method=None, threads=None, **kwargs):
        if method is None:
            methods = Configuration().fourier.fft.method
        else:
            methods = [method]

        if threads is None:
            # Use a single thread for small FFTs by default. The exact cutoff is very
            # rough changes from computer to computer. We choose 256x256 as a general guide.
            if x.size < 256**2:
                threads_attempts = [1]
            else:
                threads_attempts = [_CPU_COUNT, 1]

        # Try multithreaded first, but fall back upon single-threaded if that doesn't work.
        for threads in threads_attempts:
            for method in methods:
                try:
                    if method == 'mkl' and mkl_fft is not None and mkl_func is not None:
                        return mkl_func(x, *args, **kwargs, overwrite_x=overwrite_x)
                    elif method == 'fftw' and pyfftw is not None:
                        return pyfftw_func(x, *args, **kwargs, workers=threads, overwrite_x=overwrite_x)
                    elif method == 'scipy':
                        return scipy_func(x, *args, **kwargs, workers=threads, overwrite_x=overwrite_x)
                    elif method == 'numpy':
                        # Numpy doesn't always retain the correct bit depth. Cast the results to the correct bit depth here.
                        dtype_in = x.dtype
                        real_out = func_name.startswith('ir') or func_name.startswith('h')

                        if dtype_in == 'float32' or dtype_in == 'complex64':
                            dtype_out = 'float32' if real_out else 'complex64'
                        else:
                            dtype_out = 'float64' if real_out else 'complex128'

                        return numpy_func(x, *args, **kwargs).astype(dtype_out, copy=False)
                except Exception as e:
                    warnings.warn(f'Method {method} raised an exception "{e}". Falling back to other methods.')
            else:
                warnings.warn(f'No suitable/working FFT method could be found using {threads} threads.')
        else:
            raise ValueError('No suitable/working FFT method could be found.')

    return func

# Make all individual functions using this function generator.
fft = _make_func('fft')
fft2 = _make_func('fft2')
fftn = _make_func('fftn')

ifft = _make_func('ifft')
ifft2 = _make_func('ifft2')
ifftn = _make_func('ifftn')

rfft = _make_func('rfft')
rfft2 = _make_func('rfft2')
rfftn = _make_func('rfftn')

irfft = _make_func('irfft')
irfft2 = _make_func('irfft2')
irfftn = _make_func('irfftn')

hfft = _make_func('hfft')

ihfft = _make_func('ihfft')

# For completeness, we include fftshift and fftfreq functions.
fftshift = np.fft.fftshift
ifftshift = np.fft.ifftshift

fftfreq = np.fft.fftfreq
rfftfreq = np.fft.rfftfreq



import numpy as np
from scipy.sparse import csr_matrix
import numexpr as ne

try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

from .optical_element import OpticalElement
from ..field import make_uniform_grid, evaluate_supersampled
from ..mode_basis import ModeBasis, make_gaussian_pokes
from ..interpolation import make_linear_interpolator_separated
from ..util import read_fits

def make_actuator_positions(num_actuators_across_pupil, actuator_spacing, x_tilt=0, y_tilt=0, z_tilt=0):
    '''Make actuator positions using the BMC convention.

    Parameters
    ----------
    num_actuators_across_pupil : integer
        The number of actuators across the pupil. The total number of actuators will be this number squared.
    actuator_spacing : scalar
        The spacing between actuators before tilting the deformable mirror.
    x_tilt : scalar
        The tilt of the deformable mirror around the x-axis in radians.
    y_tilt : scalar
        The tilt of the deformable mirror around the y-axis in radians.
    z_tilt : scalar
        The tilt of the deformable mirror around the z-axis in radians.

    Returns
    -------
    Grid
        The actuator positions.
    '''
    extent = actuator_spacing * num_actuators_across_pupil
    grid = make_uniform_grid(num_actuators_across_pupil, [extent, extent]).scaled(np.cos([y_tilt, x_tilt]))

    if z_tilt == 0:
        return grid

    grid = grid.rotated(z_tilt)
    return grid

def make_gaussian_influence_functions(pupil_grid, num_actuators_across_pupil, actuator_spacing, crosstalk=0.15, cutoff=3, x_tilt=0, y_tilt=0, z_tilt=0, oversampling=1):
    '''Create influence functions with a Gaussian profile.

    The default value for the crosstalk is representative for Boston Micromachines DMs.

    Parameters
    ----------
    pupil_grid : Grid
        The grid on which to calculate the influence functions.
    num_actuators_across_pupil : integer
        The number of actuators across the pupil. The total number of actuators will be this number squared.
    actuator_spacing : scalar
        The spacing between actuators before tilting the deformable mirror.
    crosstalk : scalar
        The amount of crosstalk between the actuators. This is defined as the value of the influence function
        at a nearest-neighbour actuator.
    cutoff : scalar
        The distance from the center of the actuator, as a fraction of the actuator spacing, where the
        influence function is truncated to zero.
    x_tilt : scalar
        The tilt of the deformable mirror around the x-axis in radians.
    y_tilt : scalar
        The tilt of the deformable mirror around the y-axis in radians.
    z_tilt : scalar
        The tilt of the deformable mirror around the z-axis in radians.
    oversampling : integer
        The oversamping factor when creating the Gaussian. Default: 1.

    Returns
    -------
    ModeBasis
        The influence functions for each of the actuators.
    '''
    actuator_positions = make_actuator_positions(num_actuators_across_pupil, actuator_spacing)

    sigma = actuator_spacing / (np.sqrt((-2 * np.log(crosstalk))))
    cutoff = actuator_spacing / sigma * cutoff

    def transform_poke(poke):
        def new_poke(grid):
            p = poke(grid.scaled(1 / np.cos([y_tilt, x_tilt])).rotated(-z_tilt))
            p /= np.cos(x_tilt) * np.cos(y_tilt)

            return p
        return new_poke

    pokes = make_gaussian_pokes(None, actuator_positions, sigma, cutoff)
    pokes = [transform_poke(p) for p in pokes]
    pokes = evaluate_supersampled(pokes, pupil_grid, oversampling, make_sparse=True)

    return pokes

def make_xinetics_influence_functions(pupil_grid, num_actuators_across_pupil, actuator_spacing, x_tilt=0, y_tilt=0, z_tilt=0):
    '''Create influence functions for a Xinetics deformable mirror.

    This function uses a The rotation of the deformable mirror will be done in the order X-Y-Z.

    Parameters
    ----------
    pupil_grid : Grid
        The grid on which to calculate the influence functions.
    num_actuators_across_pupil : integer
        The number of actuators across the pupil. The total number of actuators will be this number squared.
    actuator_spacing : scalar
        The spacing between actuators before tilting the deformable mirror.
    x_tilt : scalar
        The tilt of the deformable mirror around the x-axis in radians.
    y_tilt : scalar
        The tilt of the deformable mirror around the y-axis in radians.
    z_tilt : scalar
        The tilt of the deformable mirror around the z-axis in radians.

    Returns
    -------
    ModeBasis
        The influence functions for each of the actuators.
    '''
    actuator_positions = make_actuator_positions(num_actuators_across_pupil, actuator_spacing)

    # Stretch and rotate pupil_grid to correct for tilted DM
    evaluated_grid = pupil_grid.scaled(1 / np.cos([y_tilt, x_tilt])).rotated(-z_tilt)

    # Read in actuator shape from file.
    f = files('hcipy.optics').joinpath('influence_dm5v2.fits')
    with f.open('rb') as fp:
        actuator = np.squeeze(read_fits(fp)).astype('float')
    actuator /= actuator.max()

    # Convert actuator into linear interpolator.
    actuator_grid = make_uniform_grid(actuator.shape, np.array(actuator.shape) * actuator_spacing / 10.0)
    actuator = make_linear_interpolator_separated(actuator.ravel(), actuator_grid, 0)

    def poke(p):
        res = csr_matrix(actuator(evaluated_grid.shifted(-p))) / np.cos(x_tilt) * np.cos(y_tilt)
        res.eliminate_zeros()

        return res

    return ModeBasis([poke(p) for p in actuator_positions.points], pupil_grid)

def find_illuminated_actuators(basis, aperture, power_cutoff=0.1):
    '''Find the illuminated modes.

    A subset of the modes is selected based on the aperture function and a power cutoff.

    Parameters
    ----------
    basis : ModeBasis
        The mode basis for which we want to find the illuminated modes.
    aperture : Field or array_like
        The aperture
    power_cutoff : scalar
        The minimal required power over the aperture.

    Returns
    -------
    ModeBasis
        The illuminated influence functions.
    '''
    total_power = np.sum(abs(basis._transformation_matrix)**2, axis=0)
    masked_power = np.sum(abs(basis._transformation_matrix[aperture > 0])**2, axis=0)
    illuminated_actuator_mask = masked_power >= (power_cutoff * total_power)

    return ModeBasis(basis._transformation_matrix[:, illuminated_actuator_mask], basis.grid)

class DeformableMirror(OpticalElement):
    '''A deformable mirror using influence functions.

    This class does not contain any temporal simulation (ie. settling time),
    and assumes that there is no crosstalk between actuators.

    Parameters
    ----------
    influence_functions : ModeBasis
        The influence function for each of the actuators.
    '''
    def __init__(self, influence_functions):
        self.influence_functions = influence_functions

        self.actuators = np.zeros(len(influence_functions))
        self._actuators_for_cached_surface = None

        self.input_grid = influence_functions.grid
        self._surface = self.input_grid.zeros()

    @property
    def num_actuators(self):
        return self._actuators.size

    @property
    def actuators(self):
        return self._actuators

    @actuators.setter
    def actuators(self, actuators):
        self._actuators = actuators

    def forward(self, wavefront):
        '''Propagate a wavefront through the deformable mirror.

        Parameters
        ----------
        wavefront : Wavefront
            The incoming wavefront.

        Returns
        -------
        Wavefront
            The reflected wavefront.
        '''
        wf = wavefront.copy()

        variables = {'alpha': 2j * wavefront.wavenumber, 'surf': self.surface}
        wf.electric_field *= ne.evaluate('exp(alpha * surf)', local_dict=variables)

        return wf

    def backward(self, wavefront):
        '''Propagate a wavefront backwards through the deformable mirror.

        Parameters
        ----------
        wavefront : Wavefront
            The incoming wavefront.

        Returns
        -------
        Wavefront
            The reflected wavefront.
        '''
        wf = wavefront.copy()

        variables = {'alpha': -2j * wavefront.wavenumber, 'surf': self.surface}
        wf.electric_field *= ne.evaluate('exp(alpha * surf)', local_dict=variables)

        return wf

    @property
    def influence_functions(self):
        '''The influence function for each of the actuators of this deformable mirror.
        '''
        return self._influence_functions

    @influence_functions.setter
    def influence_functions(self, influence_functions):
        self._influence_functions = influence_functions
        self._actuators_for_cached_surface = None

    @property
    def surface(self):
        '''The surface of the deformable mirror in meters.
        '''
        if self._actuators_for_cached_surface is not None:
            if np.all(self.actuators == self._actuators_for_cached_surface):
                return self._surface

        self._surface = self.influence_functions.linear_combination(self.actuators)
        self._actuators_for_cached_surface = self.actuators.copy()

        return self._surface

    @property
    def opd(self):
        '''The optical path difference in meters that this deformable
        mirror induces.
        '''
        return 2 * self.surface

    def random(self, rms):
        '''Set the dm actuators with random white noise of a certain rms.

        Parameters
        ----------
        rms : scalar
            The dm surface rms.
        '''
        self._actuators = np.random.randn(self._actuators.size) * rms

    def phase_for(self, wavelength):
        '''Get the phase in radians that is added to a wavefront with a specified wavelength.

        Parameters
        ----------
        wavelength : scalar
            The wavelength at which to calculate the phase deformation.

        Returns
        -------
        Field
            The calculated phase deformation.
        '''
        return 2 * self.surface * 2 * np.pi / wavelength

    def flatten(self):
        '''Flatten the DM by setting all actuators to zero.
        '''
        self._actuators = np.zeros(len(self.influence_functions))

def label_actuator_centroid_positions(influence_functions, label_format='{:d}', **text_kwargs):
    '''Display centroid positions for a set of influence functions.

    The location of each of the actuators is calculated using a weighted centroid, and
    at that location a label is written to the open Matplotlib Figure. The text can be
    modified with `label_format`, which is formatted with new-style Python formatting:
    `label_format.format(i)` where `i` is the actuator index.

    Parameters
    ----------
    influence_functions : ModeBasis
        The influence function for each actuator.
    label_format : string
        The text that will be displayed at the actuator centroid. This must be a new-style
        formattable string.

    Raises
    ------
    ValueError
        If the influence functions mode basis does not contain a grid.
    '''
    import matplotlib.pyplot as plt

    if influence_functions.grid is None:
        raise ValueError('The influence functions mode basis must contain a grid to calcuate centroids.')

    grid = influence_functions.grid.as_('cartesian')
    x, y = grid.coords

    # Center the labels unless overridden by the user.
    kwargs = {'verticalalignment': 'center', 'horizontalalignment': 'center'}
    kwargs.update(text_kwargs)

    for i, act in enumerate(influence_functions):
        x_pos = (act * x).sum() / act.sum()
        y_pos = (act * y).sum() / act.sum()
        pos = (x_pos, y_pos)

        plt.annotate(label_format.format(i), xy=pos, **kwargs)



import numpy as np
from ..field import Field

def spherical_surface_sag(radius_of_curvature):
    '''Makes a Field generator for the surface sag of an even aspherical surface.

    Parameters
    ----------
    radius_of_curvature : scalar
        The radius of curvature of the surface.

    Returns
    -------
    Field generator
        This function can be evaluated on a grid to get the sag profile.
    '''
    return conical_surface_sag(radius_of_curvature, conic_constant=0)

def parabolic_surface_sag(radius_of_curvature):
    '''Makes a Field generator for the surface sag of an even aspherical surface.

    Parameters
    ----------
    radius_of_curvature : scalar
        The radius of curvature of the surface.

    Returns
    -------
    Field generator
        This function can be evaluated on a grid to get the sag profile.
    '''
    return conical_surface_sag(radius_of_curvature, conic_constant=-1)

def conical_surface_sag(radius_of_curvature, conic_constant=0):
    r'''Makes a Field generator for the surface sag of a conical surface.

    The surface profile is defined as:

    .. math:: z = \frac{cr^2}{1 + \sqrt{1-\left(1+k\right)c^2r^2}}

    with `z` the surface sag, `c` the curvature and `k` the conic constant.

    Parameters
    ----------
    radius_of_curvature : scalar
        The radius of curvature of the surface.
    conic_constant : scalar
        The conic constant of the surface

    Returns
    -------
    Field generator
        This function can be evaluated on a grid to get the sag profile.
    '''
    def func(grid):
        x = grid.x
        y = grid.y
        r = np.hypot(x, y)

        curvature = 1 / radius_of_curvature
        alpha = (1 + conic_constant) * curvature**2 * r**2
        sag = r**2 / (radius_of_curvature * (1 + np.sqrt(1 - alpha)))

        return Field(sag, grid)

    return func

def even_aspheric_surface_sag(radius_of_curvature, conic_constant=0, aspheric_coefficients=None):
    r'''Makes a Field generator for the surface sag of an even aspherical surface.

    The surface profile is defined as:

    .. math:: z = \frac{cr^2}{1 + \sqrt{1-\left(1+k\right)c^2r^2}} + \sum_i=0 a_i r^{2i+4}

    With `z` the surface sag, `c` the curvature, `k` the conic constant and  :math:`a_i` the even aspheric coefficients.

    It is important to note that this definition deviates from the Zemax definition of an even aspheric surface.
    In Zemax the 2nd order term is also included in the expansion,
    which is unnessary because the conic surface itself already accounts for the 2nd order term.

    Parameters
    ----------
    radius_of_curvature : scalar
        The radius of curvature of the surface.
    conic_constant : scalar
        The conic constant of the surface
    aspheric_coefficients : array_like
        Contains the high-order even aspheric coefficients.

    Returns
    -------
    Field generator
        This function can be evaluated on a grid to get the sag profile.
    '''
    if aspheric_coefficients is None:
        aspheric_coefficients = []

    def func(grid):
        x = grid.x
        y = grid.y
        r = np.hypot(x, y)

        # Start with a conic surface
        curvature = 1 / radius_of_curvature
        alpha = (1 + conic_constant) * curvature**2 * r**2
        sag = r**2 / (radius_of_curvature * (1 + np.sqrt(1 - alpha)))

        # Add aspheric coefficients
        # Only use the even modes and start at 4, because 0 is piston and 2 is the conic surface
        for ai, coef in enumerate(aspheric_coefficients):
            power_index = 4 + ai * 2
            sag += coef * r**power_index
        return Field(sag, grid)

    return func



from .wavefront_sensor import WavefrontSensorOptics, WavefrontSensorEstimator
from ..propagation import FraunhoferPropagator
from ..aperture import make_circular_aperture
from ..optics import SurfaceApodizer, Apodizer, TipTiltMirror
from ..field import make_pupil_grid, Field, CartesianGrid, UnstructuredCoords
from ..field import evaluate_supersampled

import numpy as np

class ModulatedPyramidWavefrontSensorOptics(WavefrontSensorOptics):
    '''The optical elements for a modulated pyramid wavefront sensor.

    This class supports both a slow (default) and a fast modulation method, chosen by the
    fast_modulation_method parameters. The fast method shifts the pyramid surface instead of
    applying tilts to the incoming wavefront. While the fast modulation method can be up to
    twice as fast, it needs to precompute all pyramid surfaces and will use up more memory.

    Parameters
    ----------
    pyramid_wavefront_sensor : WavefrontSensorOptics
        The pyramid wavefront sensor optics that are used.
    modulation : scalar
        The modulation radius in radians.
    num_steps : int
        The number of steps per modulation cycle.
    fast_modulation_method : boolean
        If True the fast propagation method will be used. Default is False.
    '''
    def __init__(self, pyramid_wavefront_sensor, modulation, num_steps=12, fast_modulation_method=False):
        self.modulation = modulation
        self.pyramid_wavefront_sensor = pyramid_wavefront_sensor
        self.tip_tilt_mirror = TipTiltMirror(self.pyramid_wavefront_sensor.input_grid)
        self.focal_grid = self.pyramid_wavefront_sensor.focal_grid

        self.pupil_to_focal = self.pyramid_wavefront_sensor.pupil_to_focal
        self.focal_to_pupil = self.pyramid_wavefront_sensor.focal_to_pupil
        self._fast_modulation_method = fast_modulation_method

        # Calculate the modulation positions
        theta = np.linspace(0, 2 * np.pi, num_steps, endpoint=False)
        x_modulation = modulation / 2 * np.cos(theta)
        y_modulation = modulation / 2 * np.sin(theta)

        self.modulation_positions = CartesianGrid(UnstructuredCoords((x_modulation, y_modulation)))

        separation = self.pyramid_wavefront_sensor._separation
        refractive_index = self.pyramid_wavefront_sensor._refractive_index
        wavelength_0 = self.pyramid_wavefront_sensor._wavelength_0

        # Create all the shifted pyramid surfaces
        def surface_function(temp_grid):
            return Field(-separation / (2 * (refractive_index(wavelength_0) - 1)) * (np.abs(temp_grid.x) + np.abs(temp_grid.y)), temp_grid)

        self._pyramid_apodizers = []
        for p in self.modulation_positions:
            # The factor of two is to compensate for the factor of 1/2 of the mirror surface.
            shifted_grid = self.focal_grid.shifted(2 * p)

            pyramid_surface = evaluate_supersampled(surface_function, shifted_grid, 4)
            pyramid_apodizer = SurfaceApodizer(Field(pyramid_surface, self.focal_grid), refractive_index)

            self._pyramid_apodizers.append(pyramid_apodizer)

    def forward(self, wavefront):
        '''Propagates a wavefront through the modulated pyramid wavefront sensor.

        Parameters
        ----------
        wavefront : Wavefront
            The input wavefront that will propagate through the system.

        Returns
        -------
        wf_modulated : list
            A list of wavefronts for each modulation position.
        '''
        if self._fast_modulation_method:
            wf_foc = self.pyramid_wavefront_sensor.spatial_filter(self.pupil_to_focal(wavefront))
            wf_modulated = [self.focal_to_pupil(apod(wf_foc)) for apod in self._pyramid_apodizers]
        else:
            wf_modulated = []

            for point in self.modulation_positions.points:
                self.tip_tilt_mirror.actuators = point
                modulated_wavefront = self.tip_tilt_mirror.forward(wavefront)

                wf_modulated.append(self.pyramid_wavefront_sensor.forward(modulated_wavefront))

        return wf_modulated

    def backward(self, wavefront):
        raise RuntimeError('This is a non-physical operation.')

class PyramidWavefrontSensorOptics(WavefrontSensorOptics):
    '''The optical elements for a pyramid wavefront sensor.

    Parameters
    ----------
    input_grid : Grid
        The grid on which the input wavefront is defined.
    output_grid : Grid
        The grid on which the output wavefront is defined.
    separation : scalar
        The separation between the pupils. The default takes the input grid extent as separation.
    pupil_diameter : scalar
        The diameter of the aperture.
    wavelength_0 : scalar
        The reference wavelength that determines the physical scales.
    q : scalar
        The focal plane oversampling coefficient. The default uses the minimal required sampling.
    num_airy : scalar
        The radius of the focal plane spatial filter in units of lambda/D at the reference wavelength.
    refractive_index : callable
        A callable that returns the refractive index as function of wavelength.
        The default is a refractive index of 1.5.
    '''
    def __init__(self, input_grid, output_grid, separation=None, pupil_diameter=None, wavelength_0=1, q=None, num_airy=None, refractive_index=lambda x: 1.5):
        if not input_grid.is_regular:
            raise ValueError('The input grid must be a regular grid.')

        self.input_grid = input_grid
        self.output_grid = output_grid

        if pupil_diameter is None:
            pupil_diameter = np.max(input_grid.delta * input_grid.shape)

        if separation is None:
            separation = pupil_diameter

        # Create the intermediate focal grid
        # Oversampling is necessary to see all frequencies in the output wavefront sensor plane
        # and we require at least 2 pixels per spatial resolution element for the default case.
        qmin = max((output_grid.delta * output_grid.dims) / (input_grid.delta * input_grid.dims))
        qmin = np.ceil(max(qmin, 2))

        if q is None:
            q = qmin
        elif q < qmin:
            raise ValueError('The requested focal plane sampling is too low to sufficiently sample the wavefront sensor output.')

        if num_airy is None:
            self.num_airy = np.max(input_grid.shape) / 2
        else:
            self.num_airy = num_airy

        num_pixels = 2 * int(self.num_airy * q)
        spatial_resolution = wavelength_0 / pupil_diameter
        self.focal_grid = make_pupil_grid(num_pixels, 2 * spatial_resolution * self.num_airy)

        self._separation = separation
        self._refractive_index = refractive_index
        self._wavelength_0 = wavelength_0

        # Make all the optical elements
        self.spatial_filter = Apodizer(make_circular_aperture(2 * self.num_airy * wavelength_0 / pupil_diameter)(self.focal_grid))
        pyramid_surface = -separation / (2 * (refractive_index(wavelength_0) - 1)) * (np.abs(self.focal_grid.x) + np.abs(self.focal_grid.y))
        self.pyramid = SurfaceApodizer(Field(pyramid_surface, self.focal_grid), refractive_index)

        # Make the propagators
        self.pupil_to_focal = FraunhoferPropagator(input_grid, self.focal_grid)
        self.focal_to_pupil = FraunhoferPropagator(self.focal_grid, self.output_grid)

    def forward(self, wavefront):
        '''Propagates a wavefront through the pyramid wavefront sensor.

        Parameters
        ----------
        wavefront : Wavefront
            The input wavefront that will propagate through the system.

        Returns
        -------
        wf_wfs : Wavefront
            The output wavefront.
        '''
        wf_focus = self.pupil_to_focal.forward(wavefront)
        wf_pyramid = self.pyramid.forward(self.spatial_filter.forward(wf_focus))
        wf_wfs = self.focal_to_pupil.forward(wf_pyramid)

        return wf_wfs

    def backward(self, wavefront):
        '''Propagates a wavefront backwards through the pyramid wavefront sensor.

        Parameters
        ----------
        wavefront : Wavefront
            The input wavefront that will propagate through the system.

        Returns
        -------
        wf_pupil : Wavefront
            The output wavefront.
        '''
        wf_focus = self.focal_to_pupil.backward(wavefront)
        wf_pyramid = self.pyramid.backward(self.spatial_filter.backward(wf_focus))
        wf_pupil = self.pupil_to_focal.backward(wf_pyramid)

        return wf_pupil

class PyramidWavefrontSensorEstimator(WavefrontSensorEstimator):
    '''Estimates the wavefront slopes from pyramid wavefront sensor images.

    Parameters
    ----------
    aperture : function
        A function which mask the pupils for the normalized differences.
    output_grid : Grid
        The grid on which the output of a pyramid wavefront sensor is sampled.

    Attributes
    ----------
    measurement_grid : Grid
        The grid on which the normalized differences are defined.
    pupil_mask : array_like
        A mask for the normalized differences.
    num_measurements : int
        The number of pixels in the output vector.
    '''
    def __init__(self, aperture):
        self.pupil_mask = aperture
        self.num_measurements = 2 * int(np.sum(self.pupil_mask > 0))

    def estimate(self, images):
        '''A function which estimates the wavefront slope from a pyramid image.

        Parameters
        ----------
        images - list
            A list of scalar intensity fields containing pyramid wavefront sensor images.

        Returns
        -------
        res - Field
            A field with wavefront sensor slopes.
        '''
        import warnings
        warnings.warn("This function does not work as expected and will be changed in a future update.", RuntimeWarning)

        image = images.shaped
        sub_shape = image.grid.shape // 2

        # Subpupils
        I_a = image[:sub_shape[0], :sub_shape[1]]
        I_b = image[sub_shape[0]:2 * sub_shape[0], :sub_shape[1]]
        I_c = image[sub_shape[0]:2 * sub_shape[0], sub_shape[1]:2 * sub_shape[1]]
        I_d = image[:sub_shape[0], sub_shape[1]:2 * sub_shape[1]]

        norm = I_a + I_b + I_c + I_d
        inv_norm = np.zeros_like(norm)
        inv_norm[norm != 0] = 1 / norm[norm != 0]

        I_x = (I_a + I_b - I_c - I_d) * inv_norm
        I_y = (I_a - I_b - I_c + I_d) * inv_norm

        I_x = I_x.ravel()[self.pupil_mask > 0]
        I_y = I_y.ravel()[self.pupil_mask > 0]

        res = Field([I_x, I_y], self.pupil_mask.grid)
        return res



import numpy as np

def inverse_truncated_modal(M, num_modes, svd=None):
    '''Invert `M` with `num_modes` modes.

    Modes are given by a singular value decomposition of the input matrix.
    If `svd` is given, it is used instead. Modes are ordered in descending
    order by their singular values.

    Parameters
    ----------
    M : ndarray
        The matrix to invert. This matrix must be two-dimensional.
    num_modes : integer
        The number of modes to use for the inversion. Higher-order modes
        are ignored.
    svd : SVD object
        A precalculated singular value decomposition object. This will
        be used if supplied, to avoid recalculation of an SVD.

    Returns
    -------
    ndarray
        The inverted matrix.
    '''
    if svd is None:
        from .singular_value_decomposition import SVD
        svd = SVD(M, num_modes)

    U, S, Vt = svd.svd

    U = U[:, :num_modes]
    S = S[:num_modes]
    Vt = Vt[:num_modes, :]

    return (Vt.T / S).dot(U.T)

def inverse_truncated(M, rcond=1e-15, svd=None):
    '''Invert `M` truncating the number of modes.

    Modes are given by a singular value decomposition of the input matrix.
    If `svd` is given, it is used instead. Modes are ordered in descending
    order by their singular values. All modes with a singular value lower than
    `rcond` times the maximum singular value, will be ignored.

    Parameters
    ----------
    M : ndarray
        The matrix to invert. This matrix must be two-dimensional.
    rcond : scalar
        The relative condition number of the highest-order mode that must
        be used for inversion.
    svd : SVD object
        A precalculated singular value decomposition object. This will
        be used if supplied, to avoid recalculation of an SVD.

    Returns
    -------
    ndarray
        The inverted matrix.
    '''
    if svd is None:
        from .singular_value_decomposition import SVD
        svd = SVD(M)

    U, S, Vt = svd.svd
    S_inv = np.array([1 / s if abs(s) > (rcond * S.max()) else 0 for s in S])

    return (Vt.T * S_inv).dot(U.T)

def inverse_tikhonov(M, rcond=1e-15, svd=None):
    '''Invert `M` using Tikhonov regularization.

    Tikhonov regularization will be an identity matrix, with strength
    `rcond` * the maximum singular value of the matrix.

    Parameters
    ----------
    M : ndarray
        The matrix to invert. This matrix must be two-dimensional.
    rcond : scalar
        The relative strength of the regularization. An identity matrix
        will be used for the regularization.
    svd : SVD object
        A precalculated singular value decomposition object. This will
        be used if supplied, to avoid recalculation of an SVD.

    Returns
    -------
    ndarray
        The inverted matrix.
    '''
    if svd is None:
        from .singular_value_decomposition import SVD
        svd = SVD(M)

    U, S, Vt = svd.svd
    S_inv = S / (S**2 + (rcond * S.max())**2)

    return (Vt.T * S_inv).dot(U.T)



import functools
import warnings

def deprecated(reason):  # pragma: no cover
    '''Decorator for deprecating functions.

    Parameters
    ----------
    reason : str
        The reason for deprecation. This will be printed with the warning.

    Returns
    -------
    function
        The new function that emits a DeprecationWarning upon use.
    '''
    def decorator(func):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            warnings.warn(f'{func.__name__} is deprecated. {reason}', DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        return wrapped
    return decorator

def deprecated_name_changed(new_func):  # pragma: no cover
    '''Decorator for deprecating functions with a name change.

    Parameters
    ----------
    new_func : function
        The new function with the same functionality.

    Returns
    -------
    function
        The new function that emits a DeprecationWarning upon use.
    '''
    def decorator(old_func):
        # I'm sure there is a way to use deprecated() above, but I can't figure that out right now.
        @functools.wraps(new_func)
        def wrapped(*args, **kwargs):
            message = f'{old_func.__name__} is deprecated. Its new name is {new_func.__name__}.'
            warnings.warn(message, DeprecationWarning, stacklevel=2)

            return new_func(*args, **kwargs)

        return wrapped
    return decorator
